package com.middlemind.Eds3;

import com.middlemind.Eds3.mock.MockGenDungeon;
import com.middlemind.Eds3.unity.GameObjectManager;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import com.middlemind.Eds3.unity.GameObjectWrapper;
import com.middlemind.Eds3.unity.GameObjectWrapperDefault;

/**
 *
 * @author Victor G. Brusca, Middlemind Games
 */
public class Eds3Client implements Runnable {

    // <editor-fold defaultstate="collapsed" desc="enumerations">
    //ENUMERATIONS
    public enum Eds3ClientMode {
        CLIENT_SEND_REQUEST_DUNGEN_SEED,
        CLIENT_RECEIVE_REQUESTED_DUNGEN_SEED,
        CLIENT_SEND_REQUEST_SPAWN_ID,
        CLIENT_RECEIVE_REQUESTED_SPAWN_ID,
        IN_LOBBY,
        WAITING,
        PREV_MODE_EMPTY,
        MODE_EMPTY,
        CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1,
        CLIENT_RECEIVE_REQUESTED_OBJ_PUBLISH_L1,
        RESTING,
        CLIENT_SEND_REQUEST_DROP_CLIENT,
        CLIENT_RECEIVE_REQUESTED_DROP_CLIENT,
        CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2,
        CLIENT_RECEIVE_REQUESTED_OBJ_PUBLISH_L2,
        GENERAL_SEND_REQUEST_1,
        GENERAL_RECEIVE_REQUESTED_1,
        GENERAL_SEND_REQUEST_2,
        GENERAL_RECEIVE_REQUESTED_2,
        GENERAL_SEND_REQUEST_3,
        GENERAL_RECEIVE_REQUESTED_3,
    }
    
    public enum Eds3ClientObjSerMode {
        L1,
        L2
    }
    // </editor-fold>
    
    // <editor-fold defaultstate="collapsed" desc="static_vars">
    //STATIC VARS
    public static final int MAX_GLOBAL_ERRORS = 25;
    public static final int MAX_ERRORS = 25;
    public static final int MAX_MOCK_MESSAGES_OUT = 25;
    public static final int DEFAULT_IN_BUFFER = 9000;
    public static final int DEFAULT_OUT_BUFFER = 9000;
    public static boolean MOCK_MODE_ON = true;
    public static boolean ASYNC_ON = false;
    public static boolean NOT_UNITY = false;
    public static int TIMING_MS = 80;
    public static boolean TIMING_MS_ON = true;
    private static int CLIENT_ID = 0;
    public static boolean CONSOLIDATE_LOGS_ON = true;
    public static int CONSOLIDATE_LOGS_LEN = 128;
    public static Eds3ClientObjSerMode OBJ_SER_MODE = Eds3ClientObjSerMode.L2;
    public static int SO_TIMEOUT_MS = 80;
    public static boolean FORCE_FLUSH = false;
    public static boolean SERVER_DRIVEN_MODE_ON = true;
    public static boolean WAIT_FOR_DROP_CLIENT_RESPONSE = false;
    // </editor-fold>

    // <editor-fold defaultstate="collapsed" desc="local_vars">
    //LOCAL VARS    
    private int totalDataIn = 0;
    private int totalMsgsIn = 0;
    private int totalDataOut = 0;
    private int totalMsgsOut = 0;

    private boolean debugOn = true;
    private DatagramSocket clientSocket = null;
    private InetSocketAddress endPoint = null;
    private int port = -1;
    private String host = "";
    private Eds3ClientMode mode;
    private Eds3ClientMode prevMode;
    private Eds3ClientMode rootMode;
    private String title = "Eds3Client";
    private boolean shuttingDown = false;
    private String version = "0101";
    private String serverVersion = null;
    private int inBufferSize = DEFAULT_IN_BUFFER;
    private int outBufferSize = DEFAULT_OUT_BUFFER;
    private int dunGenSeed = -1;
    private int spawnId = -1;
    private MockGenDungeon mgd = null;
    private int errorCount = 0;
    private int clientId = 0;
    private HandleEds3ClientComm commH = null;
    private boolean forceNextMode = false;
    private Eds3ClientMode nextMode = Eds3ClientMode.IN_LOBBY;
    private HandleDebugOut debugOut = null;
    private int globalErrorCount = 0;
    private HandleGameObjectManager gomH = null;
    private boolean paused = false;
    private String consolidatedLogs = "";
    private boolean startShuttingDown = false;
    private GameObjectWrapper gowStatic = null;
    private long sync1Sent = 0;
    private long sync1Recv = 0;
    private long sync1Diff = 0;
    private long sync1Tmp = 0;
    private long sync2Sent = 0;
    private long sync2Recv = 0;
    private long sync2Diff = 0;
    private long sync2Tmp = 0;
    private long syncFinalDiff = 0;
    // </editor-fold>

    // <editor-fold defaultstate="collapsed" desc="constructors">
    //CONSTRUCTORS
    public Eds3Client(String Host, int Port) {
        this.PrintEds3ClientModeValues();
        CLIENT_ID++;
        this.clientId = CLIENT_ID;
        this.host = Host;
        this.port = Port;
        this.mode = Eds3ClientMode.CLIENT_SEND_REQUEST_DUNGEN_SEED;
        this.prevMode = Eds3ClientMode.PREV_MODE_EMPTY;
        this.gowStatic = this.GetEmptyGameObjectWrapper();
        System.out.println(this.clientId + ": Host:" + this.host + " Port: " + this.port + " Title: " + this.title);
        System.out.println(this.clientId + ": Version: " + this.version);
    }

    public Eds3Client(String Host, int Port, String Title) {
        this.PrintEds3ClientModeValues();
        CLIENT_ID++;
        this.clientId = CLIENT_ID;
        this.host = Host;
        this.port = Port;
        this.title = Title;
        this.mode = Eds3ClientMode.CLIENT_SEND_REQUEST_DUNGEN_SEED;
        this.prevMode = Eds3ClientMode.PREV_MODE_EMPTY;
        this.gowStatic = this.GetEmptyGameObjectWrapper();
        System.out.println(this.clientId + ": Host:" + this.host + " Port: " + this.port + " Title: " + this.title);
        System.out.println(this.clientId + ": Version: " + this.version);
    }

    public Eds3Client(String Host, int Port, String Title, HandleEds3ClientComm CommH) {
        this.PrintEds3ClientModeValues();
        CLIENT_ID++;
        this.clientId = CLIENT_ID;
        this.host = Host;
        this.port = Port;
        this.title = Title;
        this.commH = CommH;
        this.mode = Eds3ClientMode.CLIENT_SEND_REQUEST_DUNGEN_SEED;
        this.prevMode = Eds3ClientMode.PREV_MODE_EMPTY;
        this.gowStatic = this.GetEmptyGameObjectWrapper();

        if (this.commH != null) {
            this.commH.HandleEds3ClientInit(this);
        }

        System.out.println(this.clientId + ": Host:" + this.host + " Port: " + this.port + " Title: " + this.title);
        System.out.println(this.clientId + ": Version: " + this.version);
    }

    public Eds3Client(String Host, int Port, String Title, HandleEds3ClientComm CommH, HandleGameObjectManager GomH) {
        this.PrintEds3ClientModeValues();
        CLIENT_ID++;
        this.clientId = CLIENT_ID;
        this.host = Host;
        this.port = Port;
        this.title = Title;
        this.commH = CommH;
        this.gomH = GomH;
        this.mode = Eds3ClientMode.CLIENT_SEND_REQUEST_DUNGEN_SEED;
        this.prevMode = Eds3ClientMode.PREV_MODE_EMPTY;
        this.gowStatic = this.GetEmptyGameObjectWrapper();

        if (this.commH != null) {
            this.commH.HandleEds3ClientInit(this);
        }

        System.out.println(this.clientId + ": Host:" + this.host + " Port: " + this.port + " Title: " + this.title);
        System.out.println(this.clientId + ": Version: " + this.version);
    }
    // </editor-fold>

    // <editor-fold defaultstate="collapsed" desc="getters_and_setters">
    //GETTERS AND SETTERS
    public void SetLocalIpAddress(String ip) {
        this.host = ip;
    }

    public long GetSyncFinalDiff()
    {
        return this.syncFinalDiff;
    }
    
    public int GetPort() {
        return port;
    }

    public void SetPort(int p) {
        this.port = p;
    }

    public void SetPort(String p) {
        try {
            int tmp = Integer.parseInt(p);
            this.port = tmp;
        } catch (Exception e) {
            this.PrintStackTrace(e.getStackTrace());
        }
    }

    public boolean IsPaused() {
        return paused;
    }

    public void SetPaused(boolean paused) {
        this.paused = paused;
    }

    public boolean IsDebugOn() {
        return this.debugOn;
    }

    public void SetDebugOn(boolean debugOn) {
        this.debugOn = debugOn;
    }

    public String GetServerVersion() {
        return this.serverVersion;
    }

    public int GetClientId() {
        return this.clientId;
    }

    public int GetSpawnId() {
        return this.spawnId;
    }

    public int GetDunGenSeed() {
        return this.dunGenSeed;
    }

    public String GetVersion() {
        return this.version;
    }

    public HandleEds3ClientComm GetCommHandler() {
        return this.commH;
    }

    public void SetCommHandler(HandleEds3ClientComm CommH) {
        this.commH = CommH;
    }

    public synchronized Eds3ClientMode GetMode() {
        return this.mode;
    }

    public synchronized void SetMode(Eds3ClientMode mode) {
        if (this.forceNextMode) {
            this.forceNextMode = false;
            this.prevMode = this.mode;
            this.mode = this.nextMode;

        } else {
            this.prevMode = this.mode;
            this.mode = mode;

        }
    }

    public HandleDebugOut GetHandleDebugOut() {
        return this.debugOut;
    }

    public void SetHandleDebugOut(HandleDebugOut dOut) {
        this.debugOut = dOut;
    }

    public int GetGlobalErrorCount() {
        return this.globalErrorCount;
    }

    public void SetGlobalErrorCount(int gec) {
        this.globalErrorCount = gec;
    }

    public HandleGameObjectManager GetHandleGameObjectManager() {
        return this.gomH;
    }

    public void SetHandleGameObjectManager(HandleGameObjectManager gomH) {
        this.gomH = gomH;
    }

    public void SetForceNextMode(Eds3ClientMode nMode) {
        this.forceNextMode = true;
        this.nextMode = nMode;
    }
    
    public boolean IsShuttingDown() {
        return this.shuttingDown;
    }
    // </editor-fold>
    
    // <editor-fold defaultstate="collapsed" desc="utility_methods">
    public void PrintEds3ClientModeValues() {
        Eds3ClientMode[] vals = Eds3ClientMode.values();
        int len = vals.length;

        for (int i = 0; i < len; i++) {
            wr("Idx: " + i + " Eds3ClientMode Value: " + vals[i].ordinal());
        }
    }
    
    public GameObjectManager GetGameObjectManager() {
        if (this.gomH != null) {
            return this.gomH.GetGameObjectManager();
        } else {
            return this.mgd.GetGameObjectManager();
        }
    }

    public GameObjectWrapper GetPlayerChar() {
        if (this.gomH != null) {
            return this.gomH.GetPlayerChar();
        } else {
            return this.mgd.GetPlayerChar();
        }
    }

    public GameObjectWrapper GetEmptyGameObjectWrapper() {
        if (this.gomH != null) {
            return this.gomH.GetEmptyGameObjectWrapper();
        } else {
            return new GameObjectWrapperDefault();
        }
    }

    public synchronized void ProcessCmd(String cmd) {
        wr("client processCmd: " + cmd);
        if (cmd != null) {
            cmd = cmd.toLowerCase();

            if (cmd.indexOf("netstats") != -1) {
                wr(this.clientId + ": TotalNetworkDataIn: " + this.totalDataIn);
                wr(this.clientId + ": TotalNetworkMsgsIn: " + this.totalMsgsIn);
                wr(this.clientId + ": TotalNetworkDataOut: " + this.totalDataOut);
                wr(this.clientId + ": TotalNetworkMsgsOut: " + this.totalMsgsOut);

            }

            if (cmd.indexOf("safeclose") != -1) {
                this.forceNextMode = true;
                this.nextMode = Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT;

            }
        }
    }

    @SuppressWarnings("CallToPrintStackTrace")
    public synchronized void wr(String s) {
        if (s == null) {
            s = "";
        }

        if (CONSOLIDATE_LOGS_ON) {
            s += "\n";
            this.consolidatedLogs += s;
            if (this.consolidatedLogs.length() >= CONSOLIDATE_LOGS_LEN) {
                PrivateWr(this.consolidatedLogs);
                this.consolidatedLogs = "";
            }
        } else {
            PrivateWr(s);
        }
    }

    private synchronized void PrivateWr(String s) {
        if (s == null) {
            s = "";
        }

        if (this.debugOn == true) {
            if (this.debugOut != null) {
                this.debugOut.wr(s);
            } else {
                if (NOT_UNITY == true) {
                    System.out.println(System.currentTimeMillis() + ": " + s);
                    if (FORCE_FLUSH) {
                        System.out.flush();
                    }
                }
            }
        }
    }

    public synchronized void ForceWr(String s) {
        this.PrivateWr(s);
    }

    public synchronized void PrintStackTrace(StackTraceElement[] st) {
        if (st != null) {
            int len = st.length;
            for (int i = 0; i < len; i++) {
                wr(st[i].toString());
            }
        }
    }
    // </editor-fold>

    // <editor-fold defaultstate="collapsed" desc="start_and_stop_methods">
    public void Stop() {
        this.shuttingDown = true;
        if (CONSOLIDATE_LOGS_ON) {
            this.ForceWr(this.consolidatedLogs);
        }
        
        try {
            this.clientSocket.close();
        } catch (Exception e) {
            wr(e.getMessage());
            PrintStackTrace(e.getStackTrace());
        }
    }

    public void SafeStop() {
        this.startShuttingDown = true;
        this.SetForceNextMode(Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT);
    }
    
    public void Start() {
        wr(this.clientId + ": Starting: " + this.title + " Version: " + this.version);
        Thread trd = new Thread(this);
        trd.setDaemon(true);
        trd.start();
    }
     
    public void StartWorking() {
        try {
            wr(this.clientId + ": StartWorking");
            this.endPoint = new InetSocketAddress(this.host, this.port);
            this.clientSocket = new DatagramSocket();
            this.clientSocket.setSoTimeout(SO_TIMEOUT_MS);
            wr(this.clientId + ": Working on port: " + this.port + " Local IP: " + this.endPoint.getAddress());

            this.shuttingDown = false;
            this.startShuttingDown = false;
            this.paused = false;
            
            while (!this.shuttingDown) {
                try {
                    if (TIMING_MS_ON) {
                        Thread.sleep(TIMING_MS);
                    }

                    if (this.paused) {
                        Thread.sleep(TIMING_MS);
                    } else {
                        if (!ASYNC_ON) {
                            ProcessCurrentMode(this.clientSocket, this.endPoint);
                        } else {
                            AsyncProcessCurrentModeCallBack(this.clientSocket, this.endPoint);
                        }
                    }
                } catch (Exception e) {
                    if (!this.shuttingDown) {
                        wr(e.getMessage());
                        PrintStackTrace(e.getStackTrace());
                    }

                    this.forceNextMode = true;
                    this.nextMode = rootMode;
                    SetMode(rootMode);
                }

                if (this.shuttingDown || this.clientSocket.isClosed()) {
                    break;
                }
            }

        } catch (Exception e) {
            if (!this.shuttingDown) {
                wr(e.getMessage());
                PrintStackTrace(e.getStackTrace());
            }
        }
    }

    @Override
    public void run() {
        StartWorking();
    }
    // </editor-fold>

    // <editor-fold defaultstate="collapsed" desc="network_methods">
    @SuppressWarnings({"UnusedAssignment", "SleepWhileHoldingLock"})
    public synchronized void ProcessCurrentMode(DatagramSocket clientSocket, InetSocketAddress endPoint) {
        try {
            if (MOCK_MODE_ON) {
                long st = (long) ((Math.random() * 3000) + 1);
                wr("Sleep for " + st + " ms.");
                Thread.sleep(st);
            }

            if (this.paused) {
                Thread.sleep(TIMING_MS);
                return;
            }

            if (this.shuttingDown) {
                return;
            }

            if (this.startShuttingDown && this.GetMode() != Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT) {
                this.forceNextMode = true;
                this.nextMode = Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT;
                this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT);
            }

            byte[] receiveData = new byte[inBufferSize];
            InetAddress ipAddress = endPoint.getAddress();
            String ip = ipAddress.toString();
            Eds3ClientMode lmode = this.GetMode();
            
            if (lmode == Eds3ClientMode.CLIENT_SEND_REQUEST_DUNGEN_SEED) {
                rootMode = lmode;
                wr(this.clientId + ": " + ip + ":" + port + ": EnterMode: CLIENT_SEND_REQUEST_DUNGEN_SEED");
                wr(this.clientId + ": " + ip + ":" + port + ": Sending SERVER_RECEIVE_REQUEST_DUNGEN_SEED msg.");
                byte[] msgId = new byte[]{Converter.toByte(Eds3Server.Eds3ServerMode.SERVER_RECEIVE_REQUEST_DUNGEN_SEED.ordinal())[3]};
                byte[] msgVer = this.version.getBytes("ASCII");
                byte[] sentOn = Converter.toByte(System.currentTimeMillis());
                int totalLen = msgId.length + msgVer.length + sentOn.length;
                sync1Sent = System.currentTimeMillis();

                byte[] dataOut = new byte[totalLen];
                System.arraycopy(msgId, 0, dataOut, 0, msgId.length);
                System.arraycopy(msgVer, 0, dataOut, msgId.length, msgVer.length);
                System.arraycopy(sentOn, 0, dataOut, (msgId.length + msgVer.length), sentOn.length);

                DatagramPacket sendPacket = new DatagramPacket(dataOut, dataOut.length, endPoint.getAddress(), port);
                NetworkWriteCallBack(sendPacket);
                this.prevMode = mode;
                this.mode = Eds3ClientMode.CLIENT_RECEIVE_REQUESTED_DUNGEN_SEED;
                wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_RECEIVE_REQUESTED_DUNGEN_SEED");

            } else if (lmode == Eds3ClientMode.CLIENT_RECEIVE_REQUESTED_DUNGEN_SEED) {
                wr(this.clientId + ": " + ip + ":" + port + ": EnterMode: CLIENT_RECEIVE_REQUESTED_DUNGEN_SEED");
                DatagramPacket receivePacket = new DatagramPacket(receiveData, receiveData.length);
                clientSocket.receive(receivePacket);

                this.totalDataIn += receivePacket.getLength();
                this.totalMsgsIn++;

                byte[] dataIn = receivePacket.getData();
                int msgId = Eds3Server.Eds3ServerMode.SERVER_SEND_REQUESTED_DUNGEN_SEED.ordinal();
                int msgIdIn = (int) dataIn[0];

                if (msgIdIn == Eds3Server.Eds3ServerMode.SERVER_SEND_SET_MODE_CMD.ordinal()) {
                    wr(this.clientId + ": " + ip + ":" + port + ": Receiving SERVER_SEND_SET_MODE_CMD");
                    int modeIdIn = (int) dataIn[1];
                    Eds3ClientMode[] vals = Eds3Client.Eds3ClientMode.values();
                    int len = vals.length;
                    
                    for (int i = 0; i < len; i++) {
                        if (vals[i].ordinal() == modeIdIn) {
                            this.mode = vals[i];
                            wr(this.clientId + ": " + ip + ":" + port + ": Setting Mode: " + this.mode);
                            break;
                        }
                    }              
                    return;
                }

                if (msgId == msgIdIn) {
                    wr(this.clientId + ": " + ip + ":" + port + ": Receiving SERVER_SEND_REQUESTED_DUNGEN_SEED msg.");
                    String version = new String(new byte[]{dataIn[1], dataIn[2], dataIn[3], dataIn[4]}, "ASCII");
                    int dunGenSeed = Converter.toInt(new byte[]{dataIn[5], dataIn[6], dataIn[7], dataIn[8]});
                    long iSentOn = Converter.toLong(new byte[]{dataIn[9], dataIn[10], dataIn[11], dataIn[12], dataIn[13], dataIn[14], dataIn[15], dataIn[16]});
                    long recvOn = System.currentTimeMillis();
                    this.serverVersion = version;
                    this.dunGenSeed = dunGenSeed;
                    
                    sync1Recv = System.currentTimeMillis();
                    sync1Tmp = (recvOn - iSentOn);
                    sync1Diff = (sync1Recv - sync1Sent);
                    
                    if(sync1Tmp > 0 && sync1Tmp > sync1Diff) {
                        sync1Diff = sync1Tmp - sync1Diff;
                    }else if(sync1Tmp < 0 && sync1Diff > sync1Tmp) {
                        sync1Diff = sync1Diff - sync1Tmp;
                    }else {
                        sync1Diff = 0;
                    }
                    
                    wr(this.clientId + ": " + ip + ":" + port + ": ServerVersion: " + version);
                    wr(this.clientId + ": " + ip + ":" + port + ": ClientVersion: " + this.version);
                    wr(this.clientId + ": " + ip + ":" + port + ": DunGenSeed: " + dunGenSeed);
                    wr(this.clientId + ": " + ip + ":" + port + ": SentOn: " + iSentOn);
                    wr(this.clientId + ": " + ip + ":" + port + ": ReceivedOn: " + recvOn);
                    wr(this.clientId + ": " + ip + ":" + port + ": Sync1Diff: " + sync1Diff);

                    if (this.commH == null) {
                        this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_SPAWN_ID);
                        wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_SPAWN_ID");
                    } else {
                        this.commH.HandleCurrentModeResult(false, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_SPAWN_ID);
                    }
                    errorCount = 0;
                } else {
                    wr(this.clientId + ": " + ip + ":" + port + ": Unexpected msg id for this mode. Exiting back to start mode for this msg chain.");
                    errorCount++;
                    if (errorCount < MAX_ERRORS) {
                        if (this.commH == null) {
                            this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_DUNGEN_SEED);
                            wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_DUNGEN_SEED");
                        } else {
                            this.commH.HandleCurrentModeResult(true, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_DUNGEN_SEED);
                        }
                    } else {
                        this.Stop();
                    }
                }

            } else if (lmode == Eds3ClientMode.CLIENT_SEND_REQUEST_SPAWN_ID) {
                rootMode = lmode;
                wr(this.clientId + ": " + ip + ":" + port + ": EnterMode: CLIENT_SEND_REQUEST_SPAWN_ID");
                wr(this.clientId + ": " + ip + ":" + port + ": Sending SERVER_RECEIVE_REQUEST_SPAWN_ID msg.");
                byte[] msgId = new byte[]{Converter.toByte(Eds3Server.Eds3ServerMode.SERVER_RECEIVE_REQUEST_SPAWN_ID.ordinal())[3]};
                byte[] msgVer = this.version.getBytes("ASCII");
                byte[] sentOn = Converter.toByte(System.currentTimeMillis());
                int totalLen = msgId.length + msgVer.length + sentOn.length;
                sync2Sent = System.currentTimeMillis();

                byte[] dataOut = new byte[totalLen];
                System.arraycopy(msgId, 0, dataOut, 0, msgId.length);
                System.arraycopy(msgVer, 0, dataOut, msgId.length, msgVer.length);
                System.arraycopy(sentOn, 0, dataOut, (msgId.length + msgVer.length), sentOn.length);

                DatagramPacket sendPacket = new DatagramPacket(dataOut, dataOut.length, endPoint.getAddress(), port);
                NetworkWriteCallBack(sendPacket);
                this.prevMode = mode;
                this.mode = Eds3ClientMode.CLIENT_RECEIVE_REQUESTED_SPAWN_ID;
                wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_RECEIVE_REQUESTED_SPAWN_ID");

            } else if (lmode == Eds3ClientMode.CLIENT_RECEIVE_REQUESTED_SPAWN_ID) {
                wr(this.clientId + ": " + ip + ":" + port + ": EnterMode: CLIENT_RECEIVE_REQUESTED_SPAWN_ID");
                DatagramPacket receivePacket = new DatagramPacket(receiveData, receiveData.length);
                clientSocket.receive(receivePacket);

                this.totalDataIn += receivePacket.getLength();
                this.totalMsgsIn++;

                byte[] dataIn = receivePacket.getData();
                int msgId = Eds3Server.Eds3ServerMode.SERVER_SEND_REQUESTED_SPAWN_ID.ordinal();
                int msgIdIn = (int) dataIn[0];

                if (msgIdIn == Eds3Server.Eds3ServerMode.SERVER_SEND_SET_MODE_CMD.ordinal()) {
                    wr(this.clientId + ": " + ip + ":" + port + ": Receiving SERVER_SEND_SET_MODE_CMD");
                    int modeIdIn = (int) dataIn[1];
                    Eds3ClientMode[] vals = Eds3Client.Eds3ClientMode.values();
                    int len = vals.length;

                    for (int i = 0; i < len; i++) {
                        if (vals[i].ordinal() == modeIdIn) {
                            this.mode = vals[i];
                            wr(this.clientId + ": " + ip + ":" + port + ": Setting Mode: " + this.mode);
                            break;
                        }
                    }
                    return;
                }

                if (msgId == msgIdIn) {
                    wr(this.clientId + ": " + ip + ":" + port + ": Receiving SERVER_SEND_REQUESTED_SPAWN_ID msg.");
                    String version = new String(new byte[]{dataIn[1], dataIn[2], dataIn[3], dataIn[4]}, "ASCII");
                    int spawnId = Converter.toInt(new byte[]{dataIn[5], dataIn[6], dataIn[7], dataIn[8]});
                    long iSentOn = Converter.toLong(new byte[]{dataIn[9], dataIn[10], dataIn[11], dataIn[12], dataIn[13], dataIn[14], dataIn[15], dataIn[16]});
                    long recvOn = System.currentTimeMillis();
                    this.serverVersion = version;
                    this.spawnId = spawnId;

                    if (Eds3Client.MOCK_MODE_ON) {
                        mgd = new MockGenDungeon(this.dunGenSeed, this.spawnId);
                        MockGenDungeon.SLEEP = true;
                        mgd.GenerateObjects();
                        mgd.GenerateSpawnObjects();
                    }

                    sync2Recv = System.currentTimeMillis();
                    sync2Tmp = (recvOn - iSentOn);
                    sync2Diff = (sync2Recv - sync2Sent);
                                        
                    if(sync2Tmp > 0 && sync2Tmp > sync2Diff) {
                        sync2Diff = sync2Tmp - sync2Diff;
                    }else if(sync2Tmp < 0 && sync2Diff > sync2Tmp) {
                        sync2Diff = sync2Diff - sync2Tmp;
                    }else {
                        sync2Diff = 0;
                    }
                    syncFinalDiff = (sync1Diff + sync2Diff) / 2;
                    
                    if(syncFinalDiff < 0) {
                        syncFinalDiff *= -1;
                    }

                    if(syncFinalDiff > 100) {
                        syncFinalDiff = 15;
                    }
                    
                    wr(this.clientId + ": " + ip + ":" + port + ": ServerVersion: " + version);
                    wr(this.clientId + ": " + ip + ":" + port + ": ClientVersion: " + this.version);
                    wr(this.clientId + ": " + ip + ":" + port + ": SpawnId/PlayerId: " + spawnId);
                    wr(this.clientId + ": " + ip + ":" + port + ": SentOn: " + iSentOn);
                    wr(this.clientId + ": " + ip + ":" + port + ": ReceivedOn: " + recvOn);
                    wr(this.clientId + ": " + ip + ":" + port + ": Sync1Diff: " + sync1Diff);
                    wr(this.clientId + ": " + ip + ":" + port + ": Sync2Diff: " + sync2Diff);
                    wr(this.clientId + ": " + ip + ":" + port + ": SyncFinalDiff: " + syncFinalDiff);
                    
                    if (this.commH == null) {
                        if (Eds3Client.OBJ_SER_MODE == Eds3ClientObjSerMode.L2) {
                            this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2);
                            wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2");
                        } else {
                            this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1);
                            wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1");
                        }
                    } else {
                        if (Eds3Client.OBJ_SER_MODE == Eds3ClientObjSerMode.L2) {
                            this.commH.HandleCurrentModeResult(false, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2);
                        } else {
                            this.commH.HandleCurrentModeResult(false, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1);
                        }
                    }
                    errorCount = 0;
                } else {
                    wr(this.clientId + ": " + ip + ":" + port + ": Unexpected msg id for this mode. Exiting back to start mode for this msg chain.");
                    errorCount++;
                    if (errorCount < MAX_ERRORS) {
                        if (this.commH == null) {
                            this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_SPAWN_ID);
                            wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_SPAWN_ID");
                        } else {
                            this.commH.HandleCurrentModeResult(true, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_SPAWN_ID);
                        }
                    } else {
                        this.Stop();
                    }
                }

            } else if (lmode == Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT) {
                rootMode = lmode;
                wr(this.clientId + ": " + ip + ":" + port + ": EnterMode: CLIENT_SEND_REQUEST_DROP_CLIENT");
                wr(this.clientId + ": " + ip + ":" + port + ": Sending SERVER_RECEIVE_REQUEST_DROP_CLIENT msg.");
                byte[] msgId = new byte[]{Converter.toByte(Eds3Server.Eds3ServerMode.SERVER_RECEIVE_REQUEST_DROP_CLIENT.ordinal())[3]};
                byte[] msgVer = this.version.getBytes("ASCII");
                byte[] sentOn = Converter.toByte(System.currentTimeMillis() - syncFinalDiff);
                int totalLen = msgId.length + msgVer.length + sentOn.length;

                byte[] dataOut = new byte[totalLen];
                System.arraycopy(msgId, 0, dataOut, 0, msgId.length);
                System.arraycopy(msgVer, 0, dataOut, msgId.length, msgVer.length);
                System.arraycopy(sentOn, 0, dataOut, (msgId.length + msgVer.length), sentOn.length);

                DatagramPacket sendPacket = new DatagramPacket(dataOut, dataOut.length, endPoint.getAddress(), port);
                NetworkWriteCallBack(sendPacket);
                this.prevMode = mode;
                this.mode = Eds3ClientMode.CLIENT_RECEIVE_REQUESTED_DROP_CLIENT;
                wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_DROP_CLIENT");
                
                if(!WAIT_FOR_DROP_CLIENT_RESPONSE) {
                    if (this.commH == null) {
                        this.Stop();
                    } else {
                        this.Stop();
                        this.commH.HandleSafeStopComplete();
                    }
                }

            } else if (lmode == Eds3ClientMode.CLIENT_RECEIVE_REQUESTED_DROP_CLIENT) {
                wr(this.clientId + ": " + ip + ":" + port + ": EnterMode: CLIENT_RECEIVE_REQUESTED_DROP_CLIENT");
                DatagramPacket receivePacket = new DatagramPacket(receiveData, receiveData.length);
                clientSocket.receive(receivePacket);

                this.totalDataIn += receivePacket.getLength();
                this.totalMsgsIn++;

                byte[] dataIn = receivePacket.getData();
                int msgId = Eds3Server.Eds3ServerMode.SERVER_SEND_REQUESTED_DROP_CLIENT.ordinal();
                int msgIdIn = (int) dataIn[0];

                if (msgIdIn == Eds3Server.Eds3ServerMode.SERVER_SEND_SET_MODE_CMD.ordinal()) {
                    wr(this.clientId + ": " + ip + ":" + port + ": Receiving SERVER_SEND_SET_MODE_CMD");
                    int modeIdIn = (int) dataIn[1];
                    Eds3ClientMode[] vals = Eds3Client.Eds3ClientMode.values();
                    int len = vals.length;

                    for (int i = 0; i < len; i++) {
                        if (vals[i].ordinal() == modeIdIn) {
                            this.mode = vals[i];
                            wr(this.clientId + ": " + ip + ":" + port + ": Setting Mode: " + this.mode);
                            break;
                        }
                    }
                    return;
                }

                if (msgId == msgIdIn) {
                    wr(this.clientId + ": " + ip + ":" + port + ": Receiving SERVER_SEND_REQUESTED_DROP_CLIENT msg.");
                    String version = new String(new byte[]{dataIn[1], dataIn[2], dataIn[3], dataIn[4]}, "ASCII");
                    int resId = Converter.toInt(new byte[]{dataIn[5], dataIn[6], dataIn[7], dataIn[8]});
                    long iSentOn = Converter.toLong(new byte[]{dataIn[9], dataIn[10], dataIn[11], dataIn[12], dataIn[13], dataIn[14], dataIn[15], dataIn[16]});
                    long recvOn = System.currentTimeMillis();

                    this.serverVersion = version;

                    wr(this.clientId + ": " + ip + ":" + port + ": ServerVersion: " + version);
                    wr(this.clientId + ": " + ip + ":" + port + ": ClientVersion: " + this.version);
                    wr(this.clientId + ": " + ip + ":" + port + ": SpawnId/PlayerId: " + this.spawnId);
                    wr(this.clientId + ": " + ip + ":" + port + ": ResultId: " + resId);
                    wr(this.clientId + ": " + ip + ":" + port + ": SentOn: " + iSentOn);
                    wr(this.clientId + ": " + ip + ":" + port + ": ReceivedOn: " + recvOn);
                    wr(this.clientId + ": " + ip + ":" + port + ": SyncFinalDiff: " + syncFinalDiff);

                    if (this.commH == null) {
                        this.Stop();
                    } else {
                        this.Stop();
                        this.commH.HandleSafeStopComplete();
                    }

                    errorCount = 0;
                } else {
                    wr(this.clientId + ": " + ip + ":" + port + ": Unexpected msg id for this mode. Exiting back to start mode for this msg chain.");
                    errorCount++;
                    if (errorCount < MAX_ERRORS) {
                        if (this.commH == null) {
                            this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT);
                            wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_DROP_CLIENT");
                        } else {
                            this.commH.HandleCurrentModeResult(true, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT);
                        }
                    } else {
                        this.Stop();
                    }
                }

            } else if (lmode == Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1) {
                rootMode = lmode;
                wr(this.clientId + ": " + ip + ":" + port + ": EnterMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1");
                wr(this.clientId + ": " + ip + ":" + port + ": Sending SERVER_RECEIVE_REQUEST_OBJ_PUBLISH_L1 msg.");
                byte[] msgId = new byte[]{Converter.toByte(Eds3Server.Eds3ServerMode.SERVER_RECEIVE_REQUEST_OBJ_PUBLISH_L1.ordinal())[3]};
                byte[] msgVer = this.version.getBytes("ASCII");
                byte[] sentOn = Converter.toByte(System.currentTimeMillis() - syncFinalDiff);
                byte[] pId = Converter.toByte(this.spawnId);
                byte[] objs = null;
                if (this.gomH != null) {
                    objs = this.gomH.GetGameObjectManager().SerializeNetworkGameObjects();
                } else {
                    objs = this.mgd.SerializeNetworkGameObjects();
                }
                byte[] objsLen = Converter.toByte(objs.length);

                if (objs.length == 0) {
                    wr("Receiving NO objects to serialize... clearing published list.");
                    //this.SetMode(Eds3ClientMode.WAITING);
                    //return;
                }

                wr(this.clientId + ": " + ip + ":" + port + ": Found objsLen: " + objs.length);
                int totalLen = msgId.length + msgVer.length + sentOn.length + pId.length + objsLen.length + objs.length;
                if (totalLen > Eds3Client.DEFAULT_OUT_BUFFER) {
                    int diff = (totalLen - Eds3Client.DEFAULT_OUT_BUFFER);
                    int diffCnt = (int) (diff / (float) gowStatic.GetTotalLength());
                    wr(this.clientId + ": " + ip + ":" + port + ": Dropping " + diffCnt + " objects due to serialized object length being too long.");
                    byte[] objs2 = new byte[objs.length - (diffCnt * gowStatic.GetTotalLength())];
                    System.arraycopy(objs, 0, objs2, 0, objs2.length);
                    objs = objs2;
                    objsLen = Converter.toByte(objs.length);
                    totalLen = msgId.length + msgVer.length + sentOn.length + pId.length + objsLen.length + objs.length;
                }

                byte[] dataOut = new byte[totalLen];
                System.arraycopy(msgId, 0, dataOut, 0, msgId.length);
                System.arraycopy(msgVer, 0, dataOut, msgId.length, msgVer.length);
                System.arraycopy(sentOn, 0, dataOut, (msgId.length + msgVer.length), sentOn.length);
                System.arraycopy(pId, 0, dataOut, (msgId.length + msgVer.length + sentOn.length), pId.length);
                System.arraycopy(objsLen, 0, dataOut, (msgId.length + msgVer.length + sentOn.length + pId.length), objsLen.length);
                System.arraycopy(objs, 0, dataOut, (msgId.length + msgVer.length + sentOn.length + pId.length + objsLen.length), objs.length);

                DatagramPacket sendPacket = new DatagramPacket(dataOut, dataOut.length, endPoint.getAddress(), port);
                NetworkWriteCallBack(sendPacket);
                this.prevMode = mode;
                this.mode = Eds3ClientMode.CLIENT_RECEIVE_REQUESTED_OBJ_PUBLISH_L1;
                wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_RECEIVE_REQUEST_OBJ_PUBLISH_L1");

            } else if (lmode == Eds3ClientMode.CLIENT_RECEIVE_REQUESTED_OBJ_PUBLISH_L1) {
                wr(this.clientId + ": " + ip + ":" + port + ": EnterMode: CLIENT_RECEIVE_REQUEST_OBJ_PUBLISH_L1");
                DatagramPacket receivePacket = new DatagramPacket(receiveData, receiveData.length);
                clientSocket.receive(receivePacket);

                this.totalDataIn += receivePacket.getLength();
                this.totalMsgsIn++;

                byte[] dataIn = receivePacket.getData();
                int msgId = Eds3Server.Eds3ServerMode.SERVER_SEND_REQUESTED_OBJ_PUBLISH_L1.ordinal();
                int msgIdIn = (int) dataIn[0];

                if (msgIdIn == Eds3Server.Eds3ServerMode.SERVER_SEND_SET_MODE_CMD.ordinal()) {
                    wr(this.clientId + ": " + ip + ":" + port + ": Receiving SERVER_SEND_SET_MODE_CMD");
                    int modeIdIn = (int) dataIn[1];
                    Eds3ClientMode[] vals = Eds3Client.Eds3ClientMode.values();
                    int len = vals.length;

                    for (int i = 0; i < len; i++) {
                        if (vals[i].ordinal() == modeIdIn) {
                            this.mode = vals[i];
                            wr(this.clientId + ": " + ip + ":" + port + ": Setting Mode: " + this.mode);
                            break;
                        }
                    }
                    return;
                }

                if (msgId == msgIdIn) {
                    wr(this.clientId + ": " + ip + ":" + port + ": Receiving SERVER_SEND_REQUESTED_OBJ_PUBLISH_L1 msg.");
                    String version = new String(new byte[]{dataIn[1], dataIn[2], dataIn[3], dataIn[4]}, "ASCII");
                    long iSentOn = Converter.toLong(new byte[]{dataIn[5], dataIn[6], dataIn[7], dataIn[8], dataIn[9], dataIn[10], dataIn[11], dataIn[12]});
                    long recvOn = System.currentTimeMillis();
                    int spawnId = Converter.toInt(new byte[]{dataIn[13], dataIn[14], dataIn[15], dataIn[16]});
                    int objsLen = Converter.toInt(new byte[]{dataIn[17], dataIn[18], dataIn[19], dataIn[20]});

                    if (objsLen == 0) {
                        wr("Receiving NO objects to deserialize... exiting mode.");
                        this.SetMode(Eds3ClientMode.WAITING);
                        return;
                    }

                    wr(this.clientId + ": ServerVersion: " + version);
                    wr(this.clientId + ": ClientVersion: " + this.version);
                    wr(this.clientId + ": SpawnId/PlayerId: " + this.spawnId);
                    wr(this.clientId + ": SentOn: " + iSentOn);
                    wr(this.clientId + ": ReceivedOn: " + recvOn);
                    wr(this.clientId + ": In SpawnId/PlayerId: " + spawnId);
                    wr(this.clientId + ": Serialized Objects Length: " + objsLen);
                    wr(this.clientId + ": SyncFinalDiff: " + syncFinalDiff);
                    
                    int ttlen = objsLen / gowStatic.GetTotalLength();
                    int start = 21;
                    GameObjectWrapper gow;
                    byte[] dtmp;

                    for (int i = 0; i < ttlen; i++) {
                        gow = this.GetEmptyGameObjectWrapper();
                        dtmp = new byte[gowStatic.GetTotalLength()];
                        System.arraycopy(dataIn, start, dtmp, 0, gowStatic.GetTotalLength());
                        start += gowStatic.GetTotalLength();
                        gow.Deserialize(dtmp);

                        if (this.gomH != null) {
                            this.gomH.GetGameObjectManager().AddInNetworkGameObject(gow);
                        } else {
                            this.mgd.GetGameObjectManager().AddInNetworkGameObject(gow);
                        }

                        if (this.commH != null) {
                            this.commH.HandleDeserializeGameObjectWrapper(objsLen, spawnId, dataIn, (start - gowStatic.GetTotalLength()), ttlen, this, gow);
                        }
                    }

                    wr(this.clientId + ": Deserialized " + ttlen + " game objects.");
                    if (!SERVER_DRIVEN_MODE_ON) {
                        if (this.commH == null) {
                            this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1);
                            wr(this.clientId + ": ExitMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1");
                        } else {
                            this.commH.HandleCurrentModeResult(false, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1);
                        }
                    }
                    errorCount = 0;
                } else {
                    wr(this.clientId + ": " + ip + ":" + port + ": Unexpected msg id for this mode. Exiting back to start mode for this msg chain.");
                    errorCount++;
                    if (errorCount < MAX_ERRORS) {
                        if (this.commH == null) {
                            this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1);
                            wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1");
                        } else {
                            this.commH.HandleCurrentModeResult(true, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1);
                        }
                    } else {
                        this.Stop();
                    }
                }

            } else if (lmode == Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2) {
                rootMode = lmode;
                wr(this.clientId + ": " + ip + ":" + port + ": EnterMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2");
                wr(this.clientId + ": " + ip + ":" + port + ": Sending SERVER_RECEIVE_REQUEST_OBJ_PUBLISH_L2 msg.");
                byte[] msgId = new byte[]{Converter.toByte(Eds3Server.Eds3ServerMode.SERVER_RECEIVE_REQUEST_OBJ_PUBLISH_L2.ordinal())[3]};
                byte[] msgVer = this.version.getBytes("ASCII");
                byte[] sentOn = Converter.toByte(System.currentTimeMillis() - syncFinalDiff);
                byte[] pId = Converter.toByte(this.spawnId);
                byte[] objs = null;
                if (this.gomH != null) {
                    objs = this.gomH.GetGameObjectManager().SerializeNetworkGameObjects();
                } else {
                    objs = this.mgd.SerializeNetworkGameObjects();
                }
                byte[] objsLen = Converter.toByte(objs.length);

                if (objs.length == 0) {
                    wr("Receiving NO objects to serialize... clearing published list.");
                    //this.SetMode(Eds3ClientMode.WAITING);
                    //return;
                }

                long x = 0;
                long y = 0;
                long z = 0;
                GameObjectWrapper lgow = this.GetPlayerChar();

                if (lgow != null) {
                    x = (long) lgow.GetX();
                    y = (long) lgow.GetY();
                    z = (long) lgow.GetZ();
                }

                byte[] bX = Converter.toByte(x);
                byte[] bY = Converter.toByte(y);
                byte[] bZ = Converter.toByte(z);

                wr(this.clientId + ": " + ip + ":" + port + ": Found objsLen: " + objs.length);
                int totalLen = msgId.length + msgVer.length + sentOn.length + pId.length + bX.length + bY.length + bZ.length + objsLen.length + objs.length;
                if (totalLen > Eds3Client.DEFAULT_OUT_BUFFER) {
                    int diff = (totalLen - Eds3Client.DEFAULT_OUT_BUFFER);
                    int diffCnt = (int) (diff / (float) gowStatic.GetTotalLength());
                    wr(this.clientId + ": " + ip + ":" + port + ": Dropping " + diffCnt + " objects due to serialized object length being too long.");
                    byte[] objs2 = new byte[objs.length - (diffCnt * gowStatic.GetTotalLength())];
                    System.arraycopy(objs, 0, objs2, 0, objs2.length);
                    objs = objs2;
                    objsLen = Converter.toByte(objs.length);
                    totalLen = msgId.length + msgVer.length + sentOn.length + pId.length + bX.length + bY.length + bZ.length + objsLen.length + objs.length;
                }

                byte[] dataOut = new byte[totalLen];
                System.arraycopy(msgId, 0, dataOut, 0, msgId.length);
                System.arraycopy(msgVer, 0, dataOut, msgId.length, msgVer.length);
                System.arraycopy(sentOn, 0, dataOut, (msgId.length + msgVer.length), sentOn.length);
                System.arraycopy(pId, 0, dataOut, (msgId.length + msgVer.length + sentOn.length), pId.length);
                System.arraycopy(bX, 0, dataOut, (msgId.length + msgVer.length + sentOn.length + pId.length), bX.length);
                System.arraycopy(bY, 0, dataOut, (msgId.length + msgVer.length + sentOn.length + pId.length + bX.length), bY.length);
                System.arraycopy(bZ, 0, dataOut, (msgId.length + msgVer.length + sentOn.length + pId.length + bX.length + bY.length), bZ.length);
                System.arraycopy(objsLen, 0, dataOut, (msgId.length + msgVer.length + sentOn.length + pId.length + bX.length + bY.length + bZ.length), objsLen.length);
                System.arraycopy(objs, 0, dataOut, (msgId.length + msgVer.length + sentOn.length + pId.length + bX.length + bY.length + bZ.length + objsLen.length), objs.length);

                DatagramPacket sendPacket = new DatagramPacket(dataOut, dataOut.length, endPoint.getAddress(), port);
                NetworkWriteCallBack(sendPacket);
                this.prevMode = mode;
                this.mode = Eds3ClientMode.CLIENT_RECEIVE_REQUESTED_OBJ_PUBLISH_L2;
                wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_RECEIVE_REQUEST_OBJ_PUBLISH_L2");

            } else if (lmode == Eds3ClientMode.CLIENT_RECEIVE_REQUESTED_OBJ_PUBLISH_L2) {
                wr(this.clientId + ": " + ip + ":" + port + ": EnterMode: CLIENT_RECEIVE_REQUEST_OBJ_PUBLISH_L2");
                DatagramPacket receivePacket = new DatagramPacket(receiveData, receiveData.length);
                clientSocket.receive(receivePacket);

                this.totalDataIn += receivePacket.getLength();
                this.totalMsgsIn++;

                byte[] dataIn = receivePacket.getData();
                int msgId = Eds3Server.Eds3ServerMode.SERVER_SEND_REQUESTED_OBJ_PUBLISH_L2.ordinal();
                int msgIdIn = (int) dataIn[0];

                if (msgIdIn == Eds3Server.Eds3ServerMode.SERVER_SEND_SET_MODE_CMD.ordinal()) {
                    wr(this.clientId + ": " + ip + ":" + port + ": Receiving SERVER_SEND_SET_MODE_CMD");
                    int modeIdIn = (int) dataIn[1];
                    Eds3ClientMode[] vals = Eds3Client.Eds3ClientMode.values();
                    int len = vals.length;

                    for (int i = 0; i < len; i++) {
                        if (vals[i].ordinal() == modeIdIn) {
                            this.mode = vals[i];
                            wr(this.clientId + ": " + ip + ":" + port + ": Setting Mode: " + this.mode);
                            break;
                        }
                    }
                    return;
                }

                String version = "";
                long iSentOn = 0;
                long recvOn = 0;
                int spawnId = 0;
                byte[] bIp = null;
                String iIp = "";
                int objsLen = 0;
                int start = 0;
                int cnt = 1;
                int ttlen = 0;
                int lstart = 0;
                GameObjectWrapper gow;
                byte[] dtmp;

                if (msgId == msgIdIn) {
                    wr(this.clientId + ": " + ip + ":" + port + ": Receiving SERVER_SEND_REQUESTED_OBJ_PUBLISH_L2 msg.");

                    msgIdIn = (int) dataIn[start];
                    start += 1;

                    while (msgId == msgIdIn) {
                        version = new String(new byte[]{dataIn[start], dataIn[start + 1], dataIn[start + 2], dataIn[start + 3]}, "ASCII");
                        start += 4;

                        iSentOn = Converter.toLong(new byte[]{dataIn[start], dataIn[start + 1], dataIn[start + 2], dataIn[start + 3], dataIn[start + 4], dataIn[start + 5], dataIn[start + 6], dataIn[start + 7]});
                        start += 8;

                        recvOn = System.currentTimeMillis();
                        spawnId = Converter.toInt(new byte[]{dataIn[start], dataIn[start + 1], dataIn[start + 2], dataIn[start + 3]});
                        start += 4;

                        bIp = new byte[Eds3Server.IP_SER_LENGTH];
                        System.arraycopy(dataIn, start, bIp, 0, bIp.length);
                        iIp = new String(bIp);
                        start += bIp.length;

                        objsLen = Converter.toInt(new byte[]{dataIn[start], dataIn[start + 1], dataIn[start + 2], dataIn[start + 3]});
                        start += 4;

                        if (objsLen == 0) {
                            wr("Receiving NO objects to deserialize... exiting mode.");
                            this.SetMode(Eds3ClientMode.WAITING);
                            return;
                        }

                        wr(this.clientId + ": ServerVersion: " + version);
                        wr(this.clientId + ": ClientVersion: " + this.version);
                        wr(this.clientId + ": SpawnId/PlayerId: " + this.spawnId);
                        wr(this.clientId + ": SentOn: " + iSentOn);
                        wr(this.clientId + ": ReceivedOn: " + recvOn);
                        wr(this.clientId + ": In SpawnId/PlayerId: " + spawnId);
                        wr(this.clientId + ": In IP Address: " + iIp);
                        wr(this.clientId + ": Serialized Objects Length: " + objsLen);
                        wr(this.clientId + ": Frame: " + cnt);
                        wr(this.clientId + ": LStart: " + start);
                        wr(this.clientId + ": SyncFinalDiff: " + syncFinalDiff);
                    
                        ttlen = objsLen / gowStatic.GetTotalLength();
                        lstart = start;

                        for (int i = 0; i < ttlen; i++) {
                            gow = this.GetEmptyGameObjectWrapper();
                            dtmp = new byte[gowStatic.GetTotalLength()];

                            System.arraycopy(dataIn, lstart, dtmp, 0, gowStatic.GetTotalLength());
                            lstart += gowStatic.GetTotalLength();
                            gow.Deserialize(dtmp);

                            if (this.gomH != null) {
                                this.gomH.GetGameObjectManager().AddInNetworkGameObject(gow);
                            } else {
                                this.mgd.GetGameObjectManager().AddInNetworkGameObject(gow);
                            }

                            if (this.commH != null) {
                                this.commH.HandleDeserializeGameObjectWrapper(objsLen, spawnId, dataIn, lstart, ttlen, this, gow);
                            }
                        }

                        wr(this.clientId + ": Deserialized " + ttlen + " game objects.");
                        start += objsLen;
                        msgIdIn = (int) dataIn[start];
                        start += 1;
                        cnt += 1;
                                                
                        if(start >= dataIn.length) {
                            break;
                        }
                    }

                    if (!SERVER_DRIVEN_MODE_ON) {
                        if (this.commH == null) {
                            this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2);
                            wr(this.clientId + ": ExitMode: CLIENT_RECEIVE_REQUEST_OBJ_PUBLISH_L2");
                        } else {
                            this.commH.HandleCurrentModeResult(false, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2);
                        }
                    }
                    errorCount = 0;
                } else {
                    wr(this.clientId + ": " + ip + ":" + port + ": Unexpected msg id for this mode. Exiting back to start mode for this msg chain.");
                    errorCount++;
                    if (errorCount < MAX_ERRORS) {
                        if (this.commH == null) {
                            this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2);
                            wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2");
                        } else {
                            this.commH.HandleCurrentModeResult(true, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2);
                        }
                    } else {
                        this.Stop();
                    }
                }

            } else if (lmode == Eds3ClientMode.WAITING) {
                wr(this.clientId + ": " + ip + ":" + port + ": EnterMode: WAITING");
                wr(this.clientId + ": " + ip + ":" + port + ": Receiving WAITING msg.");

                if (Eds3Client.MOCK_MODE_ON) {
                    Thread.sleep(10);
                }

                if (this.commH == null) {
                    this.SetMode(Eds3ClientMode.RESTING);
                    wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: RESTING");
                } else {
                    this.commH.HandleCurrentModeResult(false, this, this.mode, this.prevMode, Eds3ClientMode.RESTING);
                }
                
            } else if (lmode == Eds3ClientMode.RESTING) {
                wr(this.clientId + ": " + ip + ":" + port + ": EnterMode: RESTING");
                if (MOCK_MODE_ON) {
                    Thread.sleep(1000);
                    if (this.mgd != null) {
                        this.mgd.RandomlyAdjustObjects();
                    }
                }

                if (this.commH == null) {
                    if (MOCK_MODE_ON) {
                        wr("Found total messages of " + this.totalMsgsOut + " out of " + Eds3Client.MAX_MOCK_MESSAGES_OUT);
                        if (this.totalMsgsOut >= Eds3Client.MAX_MOCK_MESSAGES_OUT) {
                            this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT);
                            wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_DROP_CLIENT");
                        } else {
                            if (Eds3Client.OBJ_SER_MODE == Eds3ClientObjSerMode.L2) {
                                this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2);
                                wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2");
                            } else {
                                this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1);
                                wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1");
                            }
                        }
                    } else {
                        if (Eds3Client.OBJ_SER_MODE == Eds3ClientObjSerMode.L1) {
                            this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1);
                            wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1");
                        } else {
                            this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2);
                            wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2");
                        }
                    }
                } else {
                    if (MOCK_MODE_ON) {
                        wr("Found total messages of " + this.totalMsgsOut + " out of " + Eds3Client.MAX_MOCK_MESSAGES_OUT);
                        if (this.totalMsgsOut >= Eds3Client.MAX_MOCK_MESSAGES_OUT) {
                            this.commH.HandleCurrentModeResult(false, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT);
                            wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_DROP_CLIENT");
                        } else {
                            if (Eds3Client.OBJ_SER_MODE == Eds3ClientObjSerMode.L2) {
                                this.commH.HandleCurrentModeResult(false, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2);
                                wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2");
                            } else {
                                this.commH.HandleCurrentModeResult(false, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1);
                                wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1");
                            }
                        }
                    } else {
                        if (Eds3Client.OBJ_SER_MODE == Eds3ClientObjSerMode.L1) {
                            this.commH.HandleCurrentModeResult(false, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1);
                            wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1");
                        } else {
                            this.commH.HandleCurrentModeResult(false, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2);
                            wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2");
                        }
                    }
                }
                
            } else {
                if (Eds3Client.MOCK_MODE_ON) {
                    Thread.sleep(10);
                }

                if (this.commH != null) {
                    DatagramPacket receivePacket = new DatagramPacket(receiveData, receiveData.length);
                    this.commH.HandleProcessModeUnknown(receivePacket, clientSocket, this);
                }
                
            }

            this.globalErrorCount = 0;
        } catch (Exception e) {
            if (!this.shuttingDown) {
                //wr("Globel exception encountered " + this.globalErrorCount + " found out of " + Eds3Client.MAX_GLOBAL_ERRORS);
                if (this.commH != null) {
                    this.commH.HandleNetworkError(e);
                }

                this.globalErrorCount++;
                if (this.globalErrorCount > Eds3Client.MAX_GLOBAL_ERRORS) {
                    this.Stop();
                }

                this.forceNextMode = true;
                this.nextMode = rootMode;
                SetMode(rootMode);

                if (e != null) {
                    if(MOCK_MODE_ON && e.getMessage().contains("Receive timed out") == false) {
                        wr(e.getMessage());
                        PrintStackTrace(e.getStackTrace());
                    }
                }
            }
        }
    }

    public synchronized void AsyncProcessCurrentModeCallBack(DatagramSocket clientSocket, InetSocketAddress endPoint) throws Exception {
        try {
            RunClientProcessCurrentModeCallBack rd = new RunClientProcessCurrentModeCallBack(clientSocket, endPoint, this);
            Thread trd = new Thread(rd);
            trd.start();
        } catch (Exception e) {
            wr(e.getMessage());
            PrintStackTrace(e.getStackTrace());
            throw e;
        }
    }

    private synchronized void AsyncNetworkReadCallBack(DatagramPacket receivePacket) throws Exception {
        try {
            RunClientReadCallBack rd = new RunClientReadCallBack(receivePacket, this);
            Thread trd = new Thread(rd);
            trd.start();
        } catch (Exception e) {
            wr(e.getMessage());
            PrintStackTrace(e.getStackTrace());
            throw e;
        }
    }

    private synchronized void AsyncNetworkWriteCallBack(DatagramPacket sendPacket) throws Exception {
        try {
            RunClientWriteCallBack rd = new RunClientWriteCallBack(sendPacket, this);
            Thread trd = new Thread(rd);
            trd.start();
        } catch (Exception e) {
            wr(e.getMessage());
            PrintStackTrace(e.getStackTrace());
            throw e;
        }
    }

    public synchronized void NetworkReadCallBack(DatagramPacket receivePacket) throws Exception {
        //Retrieve the state object and the handler socket.
        wr("");
        wr("");
        wr("NetworkReadCallBack");

        try {
            if (this.shuttingDown) {
                return;
            }

            if (this.paused) {
                return;
            }
        } catch (Exception e) {
            if (!this.shuttingDown) {
                wr("Exception caught in NetworkReadCallBack.");
                wr(e.getMessage());
                PrintStackTrace(e.getStackTrace());
                throw e;
            }
        }
    }

    public synchronized void NetworkWriteCallBack(DatagramPacket sendPacket) throws Exception {
        //Retrieve the state object and the handler socket.
        wr("");
        wr("");
        wr("NetworkWriteCallBack");

        try {
            int tlen = sendPacket.getLength();

            if (this.shuttingDown) {
                return;
            }

            if (this.paused) {
                return;
            }

            wr("DataOutLen: " + tlen + " TLen: " + tlen);
            this.totalDataOut += tlen;
            this.totalMsgsOut++;
            clientSocket.send(sendPacket);
        } catch (Exception e) {
            if (!this.shuttingDown) {
                wr("Exception caught in NetworkWriteCallBack.");
                wr(e.getMessage());
                PrintStackTrace(e.getStackTrace());
                throw e;
            }
        }
    }
    // </editor-fold>
}
