package com.middlemind.Eds3;

import com.middlemind.Eds3.unity.GameObjectWrapper;
import com.middlemind.Eds3.unity.GameObjectWrapperDefault;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.HttpURLConnection;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.NetworkInterface;
import java.net.SocketTimeoutException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Set;

/**
 *
 * @author Victor G. Brusca, Middlemind Games
 */
public class Eds3Server implements Runnable {

    // <editor-fold defaultstate="collapsed" desc="static_vars">
    //STATIC VARIABLES
    public static final int DUN_GEN_SEED_ERROR = Integer.MAX_VALUE;
    public static final int SPAWN_ID_ERROR = Integer.MAX_VALUE;
    public static final String TEST_URL = "http://www.google.com";

    public static final int MIB_INT = 1048576;
    public static final int KIB_INT = 1024;
    public static final int CHECK_LOG_SIZE_TICKS = 25;

    public static final int CHECK_LOG_SIZE = 5000000;
    public static final int DEFAULT_MAX_GAME_ROOMS = 10;
    public static final int DEFAULT_MAX_GAME_ROOM_PLAYERS = 10;

    public static final boolean DEFAULT_DEBUG_ON = false;
    public static final int DEFAULT_PORT = 49986;
    public static final int DEFAULT_TIMEOUT = 15000;

    public static final int DEFAULT_CONNECTION_TIMEOUT = 15000;
    public static final String DEFAULT_TITLE = "Eds3Server";

    public static String DEFAULT_LOCAL_IP_ADDRESS = GetLocalIPAddressStat();
    public static final int DEFAULT_IN_BUFFER = 9000;
    public static final int DEFAULT_OUT_BUFFER = 9000;

    private static final int MIN_PORT = 0;
    private static final int MAX_PORT = (Short.MAX_VALUE * 2);

    private static final int MIN_TIMEOUT = 0;
    private static final int MAX_TIMEOUT = (Short.MAX_VALUE * 2);

    private static final int MAX_IN_BUFFER_LENGTH = 9000;
    private static final int MIN_IN_BUFFER_LENGTH = (KIB_INT * 1);

    private static final int MAX_OUT_BUFFER_LENGTH = 9000;
    private static final int MIN_OUT_BUFFER_LENGTH = (KIB_INT * 1);

    private static final boolean FORCE_PORT_USE = true;
    public static boolean ASYNC_ON = true;

    public static boolean OBJ_PUBLISH_L2_DIST_LIMIT_ON = false;
    public static double OBJ_PUBLISH_L2_DIST_MAX_SQR = (1000.0 * 1000.0);
    public static double OBJ_PUBLISH_L2_DIST_MAX = 1000.0;

    public static int IP_SER_LENGTH = 32;
    public static boolean FORCE_FLUSH = false;
    public static boolean SERVER_DRIVEN_MODE_ON = true;
    // </editor-fold>

    // <editor-fold defaultstate="collapsed" desc="local_vars">
    //LOCAL VARIABLES
    private int totalDataIn = 0;
    private int totalMsgsIn = 0;
    private int totalDataOut = 0;
    private int totalMsgsOut = 0;

    private int inBufferSize = DEFAULT_IN_BUFFER;
    private int outBufferSize = DEFAULT_OUT_BUFFER;
    private String version = "0101";

    private String appName = DEFAULT_TITLE;
    private String title = DEFAULT_TITLE;
    private int port = DEFAULT_PORT;

    private int timeout = DEFAULT_TIMEOUT;

    private Hashtable<String, GameRoom> players = null;
    private ArrayList<GameRoom> gameRooms = null;
    private boolean shuttingDown = false;

    private int maxGameRooms = DEFAULT_MAX_GAME_ROOMS;
    private int maxGameRoomPlayers = DEFAULT_MAX_GAME_ROOM_PLAYERS;
    private String rootDir = ".";
    private String appDir = "";

    private String configFileName = "config.txt";
    private String configFilePath = "";
    private String debugFileName = "debug.txt";

    private String debugFilePath = "";
    private boolean debugOn = DEFAULT_DEBUG_ON;
    private BufferedWriter debug = null;

    private FileWriter fw = null;
    private int wrCount = 0;
    private String localIpAddress = DEFAULT_LOCAL_IP_ADDRESS;
    private InetSocketAddress endPoint;
    private DatagramSocket serverSocket;
    private boolean closeAllClients = false;
    private HandleEds3ServerComm commH = null;
    private GameObjectWrapper gowStatic = null;
    // </editor-fold>

    // <editor-fold defaultstate="collapsed" desc="enumerations">
    public enum Eds3ServerMode {
        SERVER_RECEIVE_REQUEST_DUNGEN_SEED,
        SERVER_SEND_REQUESTED_DUNGEN_SEED,
        SERVER_RECEIVE_REQUEST_SPAWN_ID,
        SERVER_SEND_REQUESTED_SPAWN_ID,
        SERVER_RECEIVE_REQUEST_OBJ_PUBLISH_L1,
        SERVER_SEND_REQUESTED_OBJ_PUBLISH_L1,
        SERVER_SEND_SET_MODE_CMD,
        SERVER_RECEIVE_REQUEST_DROP_CLIENT,
        SERVER_SEND_REQUESTED_DROP_CLIENT,
        SERVER_RECEIVE_REQUEST_OBJ_PUBLISH_L2,
        SERVER_SEND_REQUESTED_OBJ_PUBLISH_L2
    }

    // </editor-fold>
    
    public Eds3Server() {
        appDir = rootDir + File.separator + appName;
        configFilePath = appDir + File.separator + configFileName;
        debugFilePath = appDir + File.separator + debugFileName;

        System.out.println("");
        System.out.println("App Dir: " + appDir);
        System.out.println("Config Path: " + configFilePath);
        System.out.println("Debug Path: " + debugFilePath);
        System.out.println("Version: " + version);

        LoadSettings();
        CreateDebugFile();
        PrintEds3ModeValues();

        players = new Hashtable();
        gameRooms = new ArrayList();
        this.gowStatic = this.GetEmptyGameObjectWrapper();
    }

    public Eds3Server(HandleEds3ServerComm CommH) {
        appDir = rootDir + File.separator + appName;
        configFilePath = appDir + File.separator + configFileName;
        debugFilePath = appDir + File.separator + debugFileName;
        this.commH = CommH;

        System.out.println("");
        System.out.println("App Dir: " + appDir);
        System.out.println("Config Path: " + configFilePath);
        System.out.println("Debug Path: " + debugFilePath);
        System.out.println("Version: " + version);

        LoadSettings();
        CreateDebugFile();
        PrintEds3ModeValues();

        players = new Hashtable();
        gameRooms = new ArrayList();
        this.gowStatic = this.GetEmptyGameObjectWrapper();

        if (this.commH != null) {
            this.commH.HandleEds3ServerInit(this);
        }
    }

    // <editor-fold defaultstate="collapsed" desc="utility_methods">    
    public void PrintEds3ModeValues() {
        Eds3ServerMode[] vals = Eds3ServerMode.values();
        int len = vals.length;

        for (int i = 0; i < len; i++) {
            wr("Idx: " + i + " Eds3Mode Value: " + vals[i].ordinal());
        }
    }

    //Attempts to read the expected config file and load the settings.
    //Running the service in cmd line mode shows the expected file paths and config file format.
    @SuppressWarnings({"CallToPrintStackTrace", "ConvertToStringSwitch", "UseSpecificCatch"})
    private void LoadSettings() {
        try {
            System.out.println("");
            System.out.println("Loading settings...");
            System.out.println("Looking for data directory: " + this.appDir);
            File f = new File(this.appDir);
            if (f.exists() == false) {
                System.out.println("Creating the app directory, this is where the config file should go, config.txt.");
                f.mkdirs();
            }

            System.out.println("");
            System.out.println("Looking for config file: " + this.configFilePath);
            System.out.println("Entries in the config file are as follows.");
            System.out.println("port=[" + DEFAULT_PORT + ": The desired port number.]");
            System.out.println("debug=[" + (DEFAULT_DEBUG_ON ? "1" : "0") + ": Set to 1/true for true, 0/false for false.]");
            System.out.println("connection_timeout_ms=[" + DEFAULT_TIMEOUT + ": The number of miliseconds to wait for socket timeout.]");
            System.out.println("in_buffer_size=[" + DEFAULT_IN_BUFFER + ": The number of bytes for the input connection buffer.]");
            System.out.println("out_buffer_size=[" + DEFAULT_OUT_BUFFER + ": The number of bytes for the output connection buffer.]");
            System.out.println("local_ip_address=[" + DEFAULT_LOCAL_IP_ADDRESS + ": The local IP address to bind to.]");

            System.out.println("");
            f = new File(this.configFilePath);
            if (f.exists() == true) {
                // <editor-fold defaultstate="collapsed" desc="config_exists">
                System.out.println("Found a config file.");
                System.out.println("Loading config file values...");

                BufferedReader sr = new BufferedReader(new FileReader(this.configFilePath));

                String line = sr.readLine();
                while (line != null) {
                    boolean success = false;
                    int v = 0;
                    InetAddress ip;
                    String tmp = "";
                    String[] pair = line.split("=");
                    if (pair.length == 2) {
                        String key = pair[0];
                        String val = pair[1];
                        if ("PORT".equals(key.toUpperCase())) {
                            success = false;

                            try {
                                v = Integer.parseInt(val);
                                success = true;
                            } catch (Exception e) {

                            }

                            if (success == true) {
                                if (port >= MIN_PORT && port <= MAX_PORT) {
                                    port = v;
                                    System.out.println("Found port: " + this.port);
                                } else {
                                    port = DEFAULT_PORT;
                                    System.out.println("Port outside bounds, using default timeout: " + this.port);
                                }
                            } else {
                                port = DEFAULT_PORT;
                                System.out.println("Could not parse port, using default port: " + this.port);
                            }
                        } else if ("DEBUG".equals(key.toUpperCase())) {
                            tmp = val.toUpperCase();
                            if ("1".equals(tmp) || "TRUE".equals(tmp)) {
                                this.debugOn = true;
                            } else {
                                this.debugOn = false;
                            }
                            System.out.println("Found debug: " + this.debugOn);
                        } else if ("CONNECTION_TIMEOUT_MS".equals(key.toUpperCase())) {
                            success = false;
                            try {
                                v = Integer.parseInt(val);
                                success = true;
                            } catch (Exception e) {

                            }

                            if (success == true) {
                                if (v >= MIN_TIMEOUT && v <= MAX_TIMEOUT) {
                                    timeout = v;
                                    System.out.println("Found timeout: " + this.timeout);
                                } else {
                                    timeout = DEFAULT_TIMEOUT;
                                    System.out.println("Timeout outside bounds, using default timeout: " + this.timeout);
                                }
                            } else {
                                timeout = DEFAULT_TIMEOUT;
                                System.out.println("Could not parse timeout, using default timeout: " + this.timeout);
                            }
                        } else if ("OUT_BUFFER_SIZE".equals(key.toUpperCase())) {
                            success = false;
                            try {
                                v = Integer.parseInt(val);
                                success = true;
                            } catch (Exception e) {

                            }

                            if (success == true) {
                                if (v >= MIN_OUT_BUFFER_LENGTH && v <= MAX_OUT_BUFFER_LENGTH) {
                                    outBufferSize = v;
                                    System.out.println("Found out buffer size: " + this.outBufferSize);
                                } else {
                                    outBufferSize = DEFAULT_OUT_BUFFER;
                                    System.out.println("Out buffer size outside bounds, using default out buffer size: " + this.outBufferSize);
                                }
                            } else {
                                outBufferSize = DEFAULT_OUT_BUFFER;
                                System.out.println("Could not parse out buffer size, using default out buffer size: " + this.outBufferSize);
                            }
                        } else if ("IN_BUFFER_SIZE".equals(key.toUpperCase())) {
                            success = false;
                            try {
                                v = Integer.parseInt(val);
                                success = true;
                            } catch (Exception e) {

                            }

                            if (success == true) {
                                if (v >= MIN_IN_BUFFER_LENGTH && v <= MAX_IN_BUFFER_LENGTH) {
                                    inBufferSize = v;
                                    System.out.println("Found in buffer size: " + this.inBufferSize);
                                } else {
                                    inBufferSize = DEFAULT_IN_BUFFER;
                                    System.out.println("In buffer size outside bounds, using default in buffer size: " + this.inBufferSize);
                                }
                            } else {
                                inBufferSize = DEFAULT_IN_BUFFER;
                                System.out.println("Could not parse in buffer size, using default in buffer size: " + this.inBufferSize);
                            }
                        } else if ("LOCAL_IP_ADDRESS".equals(key.toUpperCase())) {
                            success = false;
                            try {
                                Integer.parseInt(val.replace(".", ""));
                                ip = InetAddress.getByName(val);
                                success = true;
                            } catch (Exception e) {

                            }

                            if (success == true) {
                                localIpAddress = val;
                                System.out.println("Found local ip address: " + this.localIpAddress);
                            } else {
                                localIpAddress = DEFAULT_LOCAL_IP_ADDRESS;
                                System.out.println("Could not parse local ip address, using default local ip address: " + this.localIpAddress);
                            }
                        }
                    }
                    line = sr.readLine();
                }

                // </editor-fold>
            } else {
                System.out.println("Could not find a config file.");
                System.out.println("Loading default values...");
                port = DEFAULT_PORT;
                debugOn = DEFAULT_DEBUG_ON;
                timeout = DEFAULT_TIMEOUT;
                outBufferSize = DEFAULT_OUT_BUFFER;
                inBufferSize = DEFAULT_IN_BUFFER;
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        PrintConfig();
    }

    public boolean PrintConfig() {
        System.out.println("==========CURRENT CONFIG===========");
        System.out.println("Version: " + version);
        System.out.println("Found port: " + this.port);
        System.out.println("Found debug: " + this.debugOn);
        System.out.println("Found connection_timeout_ms: " + this.timeout + " ms");
        System.out.println("Found in_buffer_size: " + this.inBufferSize + " bytes");
        System.out.println("Found out_buffer_size: " + this.outBufferSize + " bytes");

        if (GetIsNetworkAvailable() == true) {
            GetLocalIPAddress();
            return true;
        } else {
            System.out.println("Network connection not found.");
            return false;
        }
    }

    @SuppressWarnings("CallToPrintStackTrace")
    public boolean GetIsNetworkAvailable() {
        try {
            URL url = new URL(TEST_URL);
            HttpURLConnection con = (HttpURLConnection) url.openConnection();
            con.connect();

            if (con.getResponseCode() == 200) {
                return true;
            } else {
                return false;
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        return false;
    }

    @SuppressWarnings("CallToPrintStackTrace")
    public static String GetLocalIPAddressStat() {
        try {
            return InetAddress.getLocalHost().getHostAddress();
        } catch (Exception e) {
            e.printStackTrace();
            return "0.0.0.0";
        }
    }

    @SuppressWarnings("CallToPrintStackTrace")
    public String GetLocalIPAddress() {
        try {
            Enumeration e = NetworkInterface.getNetworkInterfaces();
            while (e.hasMoreElements()) {
                NetworkInterface n = (NetworkInterface) e.nextElement();
                Enumeration ee = n.getInetAddresses();
                while (ee.hasMoreElements()) {
                    InetAddress i = (InetAddress) ee.nextElement();
                    return i.toString();
                }
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        return "0.0.0.0";
    }

    public void PrintState() {
        System.out.println("==========CURRENT STATE===========");
        System.out.println("GameRooms: " + this.gameRooms.size());
        System.out.println("TotalPlayers: " + this.CountTotalPlayers());
    }

    private int CountTotalPlayers() {
        int total = 0;
        int len = gameRooms.size();
        for (int i = 0; i < len; i++) {
            total += gameRooms.get(i).players.size();
        }
        return total;
    }

    public void PrintStackTrace(StackTraceElement[] st) {
        if (st != null) {
            int len = st.length;
            for (int i = 0; i < len; i++) {
                wr(st[i].toString());
            }
        }
    }

    @SuppressWarnings("CallToPrintStackTrace")
    public synchronized void wr(String s) {
        if (this.debugOn == true) {
            System.out.println(System.currentTimeMillis() + ": " + s);
            if (s == null) {
                s = "";
            }

            if (this.debug != null) {
                try {
                    this.debug.write(s);
                    this.debug.newLine();
                    if (FORCE_FLUSH) {
                        this.debug.flush();
                    }
                } catch (Exception e) {
                    e.printStackTrace();
                    this.debugOn = false;
                    this.debug = null;
                }
                wrCount++;
            }

            if (wrCount > 0 && wrCount % CHECK_LOG_SIZE_TICKS == 0) {
                File fInf = new File(this.debugFilePath);
                if (fInf.length() >= CHECK_LOG_SIZE) {
                    try {
                        this.debug.flush();
                        this.debug.close();
                    } catch (Exception e) {
                        e.printStackTrace();
                        this.debugOn = false;
                        this.debug = null;
                    }
                    CreateDebugFile(false);
                }
            }
        }
    }
    // </editor-fold>

    // <editor-fold defaultstate="collapsed" desc="log_file_methods">
    //Opens up a new logging file.
    private void CreateDebugFile() {
        CreateDebugFile(true);
    }

    //Opens up a new logging file with the append setting.
    @SuppressWarnings("CallToPrintStackTrace")
    private void CreateDebugFile(boolean append) {
        if (this.debugOn == true) {
            try {
                File f = new File(this.debugFilePath);
                if (f.exists() == false) {
                    f.createNewFile();
                    this.debug = new BufferedWriter(new FileWriter(this.debugFilePath, false));
                } else {

                    this.debug = new BufferedWriter(new FileWriter(this.debugFilePath, append));
                }

            } catch (Exception e) {
                e.printStackTrace();
                this.debugOn = false;
                this.debug = null;
            }
            this.wrCount = 0;
        }
    }

    private void CloseDebugFile() {
        try {
            this.debug.flush();
        } catch (Exception e) {
        }

        try {
            this.debug.close();
        } catch (Exception e) {
        }

        try {
            this.fw.flush();
        } catch (Exception e) {
        }

        try {
            this.fw.close();
        } catch (Exception e) {
        }
    }
    // </editor-fold> 

    // <editor-fold defaultstate="collapsed" desc="getter_and_setter_methods">
    public GameObjectWrapper GetEmptyGameObjectWrapper() {
        if (this.commH != null) {
            return this.commH.GetEmptyGameObjectWrapper();
        } else {
            return new GameObjectWrapperDefault();
        }
    }

    public HandleEds3ServerComm GetCommH() {
        return commH;
    }

    public void SetCommH(HandleEds3ServerComm commH) {
        this.commH = commH;
    }

    public void SetPort(int p) {
        this.port = p;
    }

    public void SetPort(String p) {
        try {
            int tmp = Integer.parseInt(p);
            this.port = tmp;
        } catch (Exception e) {
            this.PrintStackTrace(e.getStackTrace());
        }
    }

    public boolean IsShuttingDown() {
        return this.shuttingDown;
    }

    public String GetTitle() {
        return title;
    }

    public Hashtable<String, GameRoom> GetPlayers() {
        return this.players;
    }
    
    //Returns true if logging is turned on.
    public boolean GetDebugOn() {
        return debugOn;
    }

    //Returns the full path to the logging file.
    public String GetDebugFilePath() {
        return debugFilePath;
    }

    //Returns the file name of the logging file.
    public String GetDebugFileName() {
        return debugFileName;
    }

    //Returns the application name of this service.
    public String GetAppName() {
        return appName;
    }

    //Returns the port number this service is using.
    public int GetPort() {
        return port;
    }

    //Returns the full path of the root directory the service is using.
    public String GetRootDir() {
        return rootDir;
    }

    //Returns the full path of the service config/log directory.
    public String GetAppDir() {
        return appDir;
    }

    //Returns the name of the config file this application is using.
    public String GetConfigFileName() {
        return configFileName;
    }

    //Returns the full path of the expected config file.
    public String GetConfigFilePath() {
        return configFilePath;
    }

    //Returns true if this service is shutting down.
    public boolean GetShuttingDown() {
        return shuttingDown;
    }

    public String GetLocalIpAddress() {
        return this.localIpAddress;
    }

    public void SetLocalIpAddress(String ip) {
        this.localIpAddress = ip;
    }
    //</editor-fold>

    // <editor-fold defaultstate="collapsed" desc="network_helper_methods">
    private synchronized int GetNewDunGenSeed() {
        return (int) (Math.random() * Eds3Server.DUN_GEN_SEED_ERROR) - 1;
    }

    private synchronized boolean AssignGameRoom(Player p) {
        if (p == null) {
            return false;
        }

        int len = this.gameRooms.size();
        GameRoom r = null;
        boolean found = false;
        int id = 0;
        String ip = p.ip;
        int port = p.port;

        for (int i = 0; i < len; i++) {
            r = this.gameRooms.get(i);
            if (r != null && r.players.containsKey(p.ip)) {
                wr(ip + ":" + port + ": Player already exists.");
                return true;
            }

            if (r != null && r.players.size() < this.maxGameRoomPlayers && !found) {
                wr(ip + ":" + port + ": Player doesn't already exist, but room does.");
                id = i;
                found = true;
            }
        }

        if (!found) {
            if (this.gameRooms.size() < this.maxGameRooms) {
                //register a new room
                wr(ip + ":" + port + ": Register a new game room.");
                r = new GameRoom();
                r.serverVersion = this.version;
                r.dunGenSeed = this.GetNewDunGenSeed();
                r.id = this.gameRooms.size();
                r.eds3server = this;
                this.gameRooms.add(r);

                //set player info
                p.roomId = this.gameRooms.indexOf(r);
                r.players.put(p.ip, p);

                p.id = r.players.get(p.ip).hashCode();
                p.spawnId = r.GetSpawnId();

                //register player with room
                this.players.put(p.ip, r);
            } else {
                wr(ip + ":" + port + ": Max rooms reached.");
                r = null;
                p.roomId = -1;
                p.id = -1;
                p.spawnId = -1;
            }
        } else {
            //get existing room
            wr(ip + ":" + port + ": Get existing game room.");
            r = gameRooms.get(id);

            //set player info
            p.roomId = this.gameRooms.indexOf(r);
            r.players.put(p.ip, p);
            p.id = r.players.get(p.ip).hashCode();
            p.spawnId = r.GetSpawnId();

            this.players.put(p.ip, r);
        }

        if (r != null) {
            wr(ip + ":" + port + ": Assigning player to room: " + p.roomId + " player id: " + p.id);
            return true;
        } else {
            return false;
        }
    }

    private synchronized boolean IsNewPlayer(String ip) {
        if (this.players != null) {
            if (this.players.containsKey(ip)) {
                return false;
            } else {
                return true;
            }
        } else {
            this.players = new Hashtable();
            return true;
        }
    }

    private synchronized boolean IsExistingPlayer(String ip) {
        return !IsNewPlayer(ip);
    }

    public synchronized void ProcessCmd(String cmd) {
        wr("server processCmd: " + cmd);
        if (cmd != null) {
            cmd = cmd.toLowerCase().trim();

            if (cmd.equals("help")) {
                wr("q: Quit server.");
                wr("netstats: Print general network stats.");
                wr("listrooms: Lists current game rooms.");
                wr("listplayers-[game room index]: List players in the given game room index.");
                wr("listlagplayers-[game room index]: List network lag players in the given game room index.");

            } else if (cmd.equals("q")) {
                wr("Stopping server...");
                this.Stop();

            } else if (cmd.equals("netstats")) {
                wr("TotalNetworkDataIn: " + this.totalDataIn);
                wr("TotalNetworkMsgsIn: " + this.totalMsgsIn);
                wr("TotalNetworkDataOut: " + this.totalDataOut);
                wr("TotalNetworkMsgsOut: " + this.totalMsgsOut);

            } else if (cmd.equals("listrooms")) {
                int len = this.gameRooms.size();
                wr("Scanning " + len + " rooms...");
                for (int i = 0; i < len; i++) {
                    wr("RoomIdx: " + (i + 1));
                    wr(this.gameRooms.get(i).toString());
                }

            } else if (cmd.indexOf("listplayers") != -1) {
                String[] s = null;
                if (cmd.indexOf("-") != -1) {
                    s = cmd.split("-");
                    int idx = Integer.parseInt(s[1]);
                    GameRoom r = this.gameRooms.get(idx);
                    Player p = null;
                    int cnt = 0;
                    if (r != null) {
                        for (String key : r.players.keySet()) {
                            p = r.players.get(key);
                            wr("PlayerIdx: " + (cnt + 1));
                            wr(p.toString());
                            cnt++;
                        }
                    }
                } else {
                    wr("No room specified...");
                }

            } else if (cmd.indexOf("listlagplayers") != -1) {
                String[] s = null;
                if (cmd.indexOf("-") != -1) {
                    s = cmd.split("-");
                    int idx = Integer.parseInt(s[1]);
                    GameRoom r = this.gameRooms.get(idx);
                    int cnt = 0;
                    if (r != null) {
                        for (Player p : r.lagPlayers) {
                            wr("LagPlayerIdx: " + (cnt + 1));
                            wr(p.toString());
                            cnt++;
                        }
                    }
                } else {
                    wr("No room specified...");
                }

            } else {
                wr("Nothing to do...");
            }
        }
    }
    // </editor-fold>

    // <editor-fold defaultstate="collapsed" desc="network_methods">
    @Override
    public void run() {
        StartListening();
    }

    //Begin listening for connection requests.
    @SuppressWarnings("UseSpecificCatch")
    private void StartListening() {
        try {
            wr("StartListening on: " + this.localIpAddress + ":" + this.port);
            this.endPoint = new InetSocketAddress(this.localIpAddress, this.port);
            this.serverSocket = new DatagramSocket(this.port, this.endPoint.getAddress());
            this.serverSocket.setSoTimeout(this.timeout);

            wr("Listening on port: " + this.port + " Local IP: " + this.endPoint.getAddress());
            byte[] receiveData = new byte[this.inBufferSize];

            while (!this.shuttingDown) {
                //Get next message block
                try {
                    DatagramPacket receivePacket = new DatagramPacket(receiveData, receiveData.length);
                    this.serverSocket.receive(receivePacket);

                    //Handle next message block in a thread
                    if (ASYNC_ON) {
                        AsyncNetworkReadCallBack(receivePacket);
                    } else {
                        NetworkReadCallBack(receivePacket);
                    }
                } catch (SocketTimeoutException e) {
                    //ignore and keep listening
                }

                if (this.shuttingDown || this.serverSocket.isClosed()) {
                    break;
                }
            }

        } catch (Exception e) {
            if (!this.shuttingDown) {
                wr(e.getMessage());
                PrintStackTrace(e.getStackTrace());
            }
        }
    }

    //Stop the service and shut it down.
    public void Stop() {
        this.CloseAllClients(true);

        try {
            Thread.sleep(3000);
        } catch (Exception e) {
        }

        this.shuttingDown = true;

        try {
            this.serverSocket.close();
        } catch (Exception e) {
            wr(e.getMessage());
            PrintStackTrace(e.getStackTrace());
        }
    }

    //Start the service listening.
    public void Start() {
        wr("Starting: " + title + " Version: " + version);
        Thread trd = new Thread(this);
        trd.setDaemon(false);
        trd.start();
    }

    public void CloseAllClients(boolean b) {
        this.closeAllClients = b;
    }

    public void AsyncNetworkReadCallBack(DatagramPacket receivePacket) {
        try {
            RunServerReadCallBack rd = new RunServerReadCallBack(receivePacket, this);
            Thread trd = new Thread(rd);
            trd.start();
        } catch (Exception e) {
            wr(e.getMessage());
            PrintStackTrace(e.getStackTrace());
        }
    }

    public void AsyncNetworkWriteCallBack(DatagramPacket sendPacket) {
        try {
            RunServerWriteCallBack rd = new RunServerWriteCallBack(sendPacket, this);
            Thread trd = new Thread(rd);
            trd.start();
        } catch (Exception e) {
            wr(e.getMessage());
            PrintStackTrace(e.getStackTrace());
        }
    }

    public synchronized void NetworkReadCallBack(DatagramPacket receivePacket) {
        //Retrieve the state object and the handler socket.
        wr("");
        wr("");
        wr("NetworkReadCallBack called ");

        try {
            if (this.shuttingDown || this.serverSocket.isClosed()) {
                return;
            }

            this.totalDataIn += receivePacket.getLength();
            this.totalMsgsIn++;

            byte[] dataIn = receivePacket.getData();
            InetAddress ipAddress = receivePacket.getAddress();
            String ip = ipAddress.toString();
            int port = receivePacket.getPort();

            if (FORCE_PORT_USE) {
                ip += ":" + port;
            }

            Player p = null;
            GameRoom r = null;

            wr("Found request id: " + dataIn[0]);
            if (dataIn[0] == Eds3ServerMode.SERVER_RECEIVE_REQUEST_DUNGEN_SEED.ordinal()) {
                wr(ip + ":" + port + ": Receiving SERVER_RECEIVE_REQUEST_DUNGEN_SEED msg.");

                int dunGenSeed = -1;
                String version = new String(new byte[]{dataIn[1], dataIn[2], dataIn[3], dataIn[4]}, "ASCII");
                long iSentOn = Converter.toLong(new byte[]{dataIn[5], dataIn[6], dataIn[7], dataIn[8], dataIn[9], dataIn[10], dataIn[11], dataIn[12]});
                long recvOn = System.currentTimeMillis();
                MsgStats msgStats = new MsgStats(iSentOn, recvOn);
                wr(ip + ":" + port + ": XportTime: " + msgStats.diff + " ms");

                if (IsNewPlayer(ip)) {
                    p = new Player(ip);
                    p.ip = ip;
                    p.port = port;
                    p.ipAddress = ipAddress;
                    p.clientVersion = version;
                    this.AssignGameRoom(p);
                }

                try {
                    r = this.players.get(ip);
                } catch (Exception ex1) {
                    this.PrintStackTrace(ex1.getStackTrace());
                    r = null;
                }

                if (r != null) {
                    p = r.players.get(ip);
                    if (p != null) {
                        p.port = port;
                        p.ipAddress = ipAddress;
                        p.clientVersion = version;
                        dunGenSeed = r.dunGenSeed;

                        long start = System.currentTimeMillis();
                        p.lastCommTimeMs = iSentOn;
                        p.LogMsgStats(msgStats);
                        r.UpdateGameRoomStats();
                        r.UpdateLaggedInactivePlayers();
                        long stop = System.currentTimeMillis();
                        long diff = stop - start;
                        wr(ip + ":" + port + ": Player maintenance took: " + diff + " ms");
                    } else {
                        p = null;
                        dunGenSeed = DUN_GEN_SEED_ERROR;
                        wr(ip + ":" + port + ": Message chain dying...");
                    }
                } else {
                    p = null;
                    dunGenSeed = DUN_GEN_SEED_ERROR;
                    wr(ip + ":" + port + ": Message chain dying...");
                }

                wr(ip + ":" + port + ": ServerVersion: " + this.version);
                wr(ip + ":" + port + ": ClientVersion: " + version);
                wr(ip + ":" + port + ": DunGenSeed: " + dunGenSeed);
                wr(ip + ":" + port + ": SentOn: " + iSentOn);
                wr(ip + ":" + port + ": ReceivedOn: " + recvOn);

                byte[] msgId = new byte[]{Converter.toByte(Eds3ServerMode.SERVER_SEND_REQUESTED_DUNGEN_SEED.ordinal())[3]};
                byte[] msgVer = this.version.getBytes("ASCII");
                byte[] data = Converter.toByte(dunGenSeed);
                byte[] sentOn = Converter.toByte(System.currentTimeMillis());
                int totalLen = msgId.length + msgVer.length + data.length + sentOn.length;

                byte[] dataOut = new byte[totalLen];
                System.arraycopy(msgId, 0, dataOut, 0, msgId.length);
                System.arraycopy(msgVer, 0, dataOut, msgId.length, msgVer.length);
                System.arraycopy(data, 0, dataOut, (msgId.length + msgVer.length), data.length);
                System.arraycopy(sentOn, 0, dataOut, (msgId.length + msgVer.length + data.length), sentOn.length);

                wr(ip + ":" + port + ": Sending SERVER_SEND_REQUESTED_DUNGEN_SEED msg");
                DatagramPacket sendPacket = new DatagramPacket(dataOut, dataOut.length, ipAddress, port);
                NetworkWriteCallBack(sendPacket);

            } else if (dataIn[0] == Eds3ServerMode.SERVER_RECEIVE_REQUEST_SPAWN_ID.ordinal()) {
                wr(ip + ":" + port + ": Found SERVER_RECEIVE_REQUEST_SPAWN_ID msg.");

                int spawnId = -1;
                String version = new String(new byte[]{dataIn[1], dataIn[2], dataIn[3], dataIn[4]}, "ASCII");
                long iSentOn = Converter.toLong(new byte[]{dataIn[5], dataIn[6], dataIn[7], dataIn[8], dataIn[9], dataIn[10], dataIn[11], dataIn[12]});
                long recvOn = System.currentTimeMillis();
                MsgStats msgStats = new MsgStats(iSentOn, recvOn);
                wr(ip + ":" + port + ": XportTime: " + msgStats.diff + " ms");

                try {
                    r = this.players.get(ip);
                } catch (Exception ex1) {
                    this.PrintStackTrace(ex1.getStackTrace());
                    r = null;
                }

                if (r != null) {
                    p = r.players.get(ip);
                    if (p != null) {
                        p.port = port;
                        p.ipAddress = ipAddress;
                        spawnId = p.spawnId;

                        long start = System.currentTimeMillis();
                        p.lastCommTimeMs = iSentOn;
                        p.LogMsgStats(msgStats);
                        r.UpdateGameRoomStats();
                        r.UpdateLaggedInactivePlayers();
                        long stop = System.currentTimeMillis();
                        long diff = stop - start;
                        wr(ip + ":" + port + ": Player maintenance took: " + diff + " ms");
                    } else {
                        p = null;
                        spawnId = SPAWN_ID_ERROR;
                        wr(ip + ":" + port + ": Message chain dying...");
                    }
                } else {
                    p = null;
                    spawnId = SPAWN_ID_ERROR;
                    wr(ip + ":" + port + ": Message chain dying...");
                }

                wr(ip + ":" + port + ": ServerVersion: " + this.version);
                wr(ip + ":" + port + ": ClientVersion: " + version);
                wr(ip + ":" + port + ": SpawnId/PlayerId: " + spawnId);
                wr(ip + ":" + port + ": SentOn: " + iSentOn);
                wr(ip + ":" + port + ": ReceivedOn: " + recvOn);

                byte[] msgId = new byte[]{Converter.toByte(Eds3ServerMode.SERVER_SEND_REQUESTED_SPAWN_ID.ordinal())[3]};
                byte[] msgVer = this.version.getBytes("ASCII");
                byte[] data = Converter.toByte(spawnId);
                byte[] sentOn = Converter.toByte(System.currentTimeMillis());
                int totalLen = msgId.length + msgVer.length + data.length + sentOn.length;

                byte[] dataOut = new byte[totalLen];
                System.arraycopy(msgId, 0, dataOut, 0, msgId.length);
                System.arraycopy(msgVer, 0, dataOut, msgId.length, msgVer.length);
                System.arraycopy(data, 0, dataOut, (msgId.length + msgVer.length), data.length);
                System.arraycopy(sentOn, 0, dataOut, (msgId.length + msgVer.length + data.length), sentOn.length);

                wr(ip + ":" + port + ": Sending SERVER_SEND_REQUESTED_SPAWN_ID msg");
                DatagramPacket sendPacket = new DatagramPacket(dataOut, dataOut.length, ipAddress, port);
                NetworkWriteCallBack(sendPacket);

            } else if (dataIn[0] == Eds3ServerMode.SERVER_RECEIVE_REQUEST_DROP_CLIENT.ordinal()) {
                wr(ip + ":" + port + ": Found SERVER_RECEIVE_REQUEST_DROP_CLIENT msg.");

                int resId = -1;
                String version = new String(new byte[]{dataIn[1], dataIn[2], dataIn[3], dataIn[4]}, "ASCII");
                long iSentOn = Converter.toLong(new byte[]{dataIn[5], dataIn[6], dataIn[7], dataIn[8], dataIn[9], dataIn[10], dataIn[11], dataIn[12]});
                long recvOn = System.currentTimeMillis();
                MsgStats msgStats = new MsgStats(iSentOn, recvOn);
                wr(ip + ":" + port + ": XportTime: " + msgStats.diff + " ms");

                try {
                    r = this.players.get(ip);
                } catch (Exception ex1) {
                    r = null;
                }

                if (r != null) {
                    p = r.players.get(ip);
                    if (p != null) {
                        p.port = port;
                        p.ipAddress = ipAddress;

                        r.players.remove(ip);
                        
                        if(r.lagPlayers.contains(p))
                        {
                            r.lagPlayers.remove(p);
                        }

                        if (r.inactivePlayers.contains(p))
                        {
                            r.inactivePlayers.remove(p);
                        }
                            
                        this.players.remove(ip);
                        if (!r.freeSpawns.contains(p.spawnId))
                        {
                            r.freeSpawns.add(p.spawnId);
                        }
                        resId = 2;

                        long start = System.currentTimeMillis();
                        //p.lastCommTimeMs = iSentOn;
                        p.lastCommTimeMs = 0;
                                                
                        p.LogMsgStats(msgStats);
                        r.UpdateGameRoomStats();
                        r.UpdateLaggedInactivePlayers();
                        
                        long stop = System.currentTimeMillis();
                        long diff = stop - start;
                        wr(ip + ":" + port + ": Player maintenance took: " + diff + " ms");
                    } else {
                        p = null;
                        this.players.remove(ip);
                        resId = 1;
                        wr(ip + ":" + port + ": Message chain dying...");
                    }
                } else {
                    p = null;
                    resId = 0;
                    wr(ip + ":" + port + ": Message chain dying...");
                }

                wr(ip + ":" + port + ": ServerVersion: " + this.version);
                wr(ip + ":" + port + ": ClientVersion: " + version);
                wr(ip + ":" + port + ": ResultId: " + resId);
                wr(ip + ":" + port + ": SentOn: " + iSentOn);
                wr(ip + ":" + port + ": ReceivedOn: " + recvOn);

                byte[] msgId = new byte[]{Converter.toByte(Eds3ServerMode.SERVER_SEND_REQUESTED_DROP_CLIENT.ordinal())[3]};
                byte[] msgVer = this.version.getBytes("ASCII");
                byte[] data = Converter.toByte(resId);
                byte[] sentOn = Converter.toByte(System.currentTimeMillis());
                int totalLen = msgId.length + msgVer.length + data.length + sentOn.length;

                byte[] dataOut = new byte[totalLen];
                System.arraycopy(msgId, 0, dataOut, 0, msgId.length);
                System.arraycopy(msgVer, 0, dataOut, msgId.length, msgVer.length);
                System.arraycopy(data, 0, dataOut, (msgId.length + msgVer.length), data.length);
                System.arraycopy(sentOn, 0, dataOut, (msgId.length + msgVer.length + data.length), sentOn.length);

                wr(ip + ":" + port + ": Sending SERVER_SEND_REQUESTED_DROP_CLIENT msg");
                DatagramPacket sendPacket = new DatagramPacket(dataOut, dataOut.length, ipAddress, port);
                NetworkWriteCallBack(sendPacket);

            } else if (dataIn[0] == Eds3ServerMode.SERVER_RECEIVE_REQUEST_OBJ_PUBLISH_L1.ordinal()) {
                wr(ip + ":" + port + ": Found SERVER_RECEIVE_REQUEST_OBJ_PUBLISH_L1 msg.");

                String version = new String(new byte[]{dataIn[1], dataIn[2], dataIn[3], dataIn[4]}, "ASCII");
                long iSentOn = Converter.toLong(new byte[]{dataIn[5], dataIn[6], dataIn[7], dataIn[8], dataIn[9], dataIn[10], dataIn[11], dataIn[12]});
                long recvOn = System.currentTimeMillis();
                MsgStats msgStats = new MsgStats(iSentOn, recvOn);
                wr(ip + ":" + port + ": XportTime: " + msgStats.diff + " ms");

                int playerId = Converter.toInt(new byte[]{dataIn[13], dataIn[14], dataIn[15], dataIn[16]});
                int objsLen = Converter.toInt(new byte[]{dataIn[17], dataIn[18], dataIn[19], dataIn[20]});
                byte[] objs = new byte[objsLen];
                
                byte[] dataOut;
                byte b1;
                byte b2;
                DatagramPacket sendPacket;
                
                try {
                    System.arraycopy(dataIn, 21, objs, 0, objs.length);
                }catch(Exception ex1) {
                    b1 = Converter.toByte(Eds3ServerMode.SERVER_SEND_SET_MODE_CMD.ordinal())[3];

                    if (this.closeAllClients) {
                        wr(ip + ":" + port + ": Sending SERVER_SEND_SET_MODE_CMD msg to CLIENT_SEND_REQUEST_DROP_CLIENT");
                        b2 = Converter.toByte(Eds3Client.Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT.ordinal())[3];
                    } else {
                        wr(ip + ":" + port + ": Sending SERVER_SEND_SET_MODE_CMD msg to WAITING");
                        b2 = Converter.toByte(Eds3Client.Eds3ClientMode.WAITING.ordinal())[3];
                    }
                    dataOut = new byte[]{b1, b2};

                    sendPacket = new DatagramPacket(dataOut, dataOut.length, ipAddress, port);
                    NetworkWriteCallBack(sendPacket);                    
                }

                try {
                    r = this.players.get(ip);
                } catch (Exception ex1) {
                    this.PrintStackTrace(ex1.getStackTrace());
                    r = null;
                }

                if (r != null) {
                    p = r.players.get(ip);
                    if (p != null) {
                        p.port = port;
                        p.ipAddress = ipAddress;
                        p.objPublish = objs;

                        long start = System.currentTimeMillis();
                        p.lastCommTimeMs = iSentOn;
                        p.LogMsgStats(msgStats);
                        r.UpdateGameRoomStats();
                        r.UpdateLaggedInactivePlayers();
                        long stop = System.currentTimeMillis();
                        long diff = stop - start;
                        wr(ip + ":" + port + ": Player maintenance took: " + diff + " ms");
                    } else {
                        p = null;
                        wr(ip + ":" + port + ": Message chain dying...");
                    }
                } else {
                    p = null;
                    wr(ip + ":" + port + ": Message chain dying...");
                }
                
                if (r != null && p != null) {
                    if(r.inactivePlayers.size() > 0) {
                        Player iap = r.inactivePlayers.get(0);
                        GameObjectWrapperDefault iapGow = new GameObjectWrapperDefault();
                        iapGow.SetPlayerId(iap.spawnId);
                        iapGow.SetMode(GameObjectWrapperDefault.GameObjectWrapperMode.DESTROY_INACTIVE_PLAYER_CHAR.ordinal());
                        iapGow.SetId(iap.id);
                        byte[] iapB = iapGow.Serialize();
                        
                        byte[] nop = null;
                        Set<String> pks = r.players.keySet();
                        Player tiap;
                        for(String pk: pks) {
                            tiap = r.players.get(pk);
                            if(tiap.objPublish == null) {
                                tiap.objPublish = iapB;
                            }else {
                                nop = new byte[tiap.objPublish.length + iapB.length];
                                System.arraycopy(tiap.objPublish, 0, nop, 0, tiap.objPublish.length);
                                System.arraycopy(iapB, 0, nop, tiap.objPublish.length, iapB.length);
                                tiap.objPublish = nop;
                            }
                        }
                    }
                    
                    wr(ip + ":" + port + ": ServerVersion: " + this.version);
                    wr(ip + ":" + port + ": ClientVersion: " + version);
                    wr(ip + ":" + port + ": SpawnId/PlayerId: " + p.spawnId);
                    wr(ip + ":" + port + ": Object Id: " + p.id);
                    wr(ip + ":" + port + ": SentOn: " + iSentOn);
                    wr(ip + ":" + port + ": ReceivedOn: " + recvOn);
                    wr(ip + ":" + port + ": In SpawnId/PlayerId: " + playerId);
                    wr(ip + ":" + port + ": Serialized Objects Length: " + objs.length);
                    wr(ip + ":" + port + ": Serialized Object Count: " + objs.length / this.gowStatic.GetTotalLength());

                    byte[] tmsgId;
                    byte[] tmsgVer;
                    byte[] tsentOn;
                    byte[] tpId;
                    byte[] tobjs;
                    byte[] tobjsLen;
                    int totalLen;
                    byte[] lb;
                    Player tp;

                    for (String key : r.players.keySet()) {
                        tp = r.players.get(key);
                        if (tp != null && tp.spawnId != p.spawnId && tp.objPublish != null) {
                            tmsgId = new byte[]{Converter.toByte(Eds3ServerMode.SERVER_SEND_REQUESTED_OBJ_PUBLISH_L1.ordinal())[3]};
                            tmsgVer = this.version.getBytes("ASCII");
                            tsentOn = Converter.toByte(System.currentTimeMillis());
                            tpId = Converter.toByte(tp.spawnId);
                            tobjs = tp.objPublish;
                            tobjsLen = Converter.toByte(tobjs.length);
                            totalLen = tmsgId.length + tmsgVer.length + tsentOn.length + tpId.length + tobjs.length + tobjsLen.length;

                            dataOut = new byte[totalLen];
                            System.arraycopy(tmsgId, 0, dataOut, 0, tmsgId.length);
                            System.arraycopy(tmsgVer, 0, dataOut, tmsgId.length, tmsgVer.length);
                            System.arraycopy(tsentOn, 0, dataOut, (tmsgId.length + tmsgVer.length), tsentOn.length);
                            System.arraycopy(tpId, 0, dataOut, (tmsgId.length + tmsgVer.length + tsentOn.length), tpId.length);
                            System.arraycopy(tobjsLen, 0, dataOut, (tmsgId.length + tmsgVer.length + tsentOn.length + tpId.length), tobjsLen.length);
                            System.arraycopy(tobjs, 0, dataOut, (tmsgId.length + tmsgVer.length + tsentOn.length + tpId.length + tobjsLen.length), tobjs.length);

                            wr(ip + ":" + port + ": Sending SERVER_SEND_REQUESTED_OBJ_PUBLISH_L1 msg for SpawnId/PlayerId: " + tp.spawnId + " to " + p.spawnId);
                            if (this.commH != null) {
                                dataOut = this.commH.HandleObjPublication(dataOut, dataOut.length, this);
                            }
                            
                            if(tp != null && tp.isInactive) {
                                wr(ip + ":" + port + ": Dropping player " + tp.spawnId + " with IP: " + tp.ip + " and ID: " + tp.id + ".");
                                r.players.remove(tp.ip);
                                this.players.remove(tp.ip);
                                
                                if(r.inactivePlayers.contains(tp)) {
                                    r.inactivePlayers.remove(tp);
                                }
                                
                                if(r.lagPlayers.contains(tp)) {
                                    r.lagPlayers.remove(tp);
                                }
                                
                                if(!r.freeSpawns.contains(p.spawnId)) {
                                    r.freeSpawns.add(p.spawnId);
                                }
                            }
                            
                            sendPacket = new DatagramPacket(dataOut, dataOut.length, ipAddress, port);
                            NetworkWriteCallBack(sendPacket);
                        } else {
                            if (tp == null) {
                                wr(ip + ":" + port + ": Skipping msg to player no player entry found.");
                            } else if (tp.spawnId == p.spawnId) {
                                wr(ip + ":" + port + ": Skipping msg to source player: " + tp.spawnId + ", " + p.spawnId);
                            } else if (tp.objPublish == null) {
                                wr(ip + ":" + port + ": Skipping msg to player no objects array to publish.");
                            }
                        }
                    }

                    if (this.commH != null) {
                        lb = this.commH.HandleProcessServerGame(System.currentTimeMillis(), this);
                        if (lb != null && lb.length > 0) {
                            wr(ip + ":" + port + ": Sending HandleProcessServerGame msg");
                            sendPacket = new DatagramPacket(lb, lb.length, ipAddress, port);
                            NetworkWriteCallBack(sendPacket);
                        }
                    }

                    if (SERVER_DRIVEN_MODE_ON) {
                        b1 = Converter.toByte(Eds3ServerMode.SERVER_SEND_SET_MODE_CMD.ordinal())[3];

                        if (this.closeAllClients) {
                            wr(ip + ":" + port + ": Sending SERVER_SEND_SET_MODE_CMD msg to CLIENT_SEND_REQUEST_DROP_CLIENT");
                            b2 = Converter.toByte(Eds3Client.Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT.ordinal())[3];
                        } else {
                            wr(ip + ":" + port + ": Sending SERVER_SEND_SET_MODE_CMD msg to WAITING");
                            b2 = Converter.toByte(Eds3Client.Eds3ClientMode.WAITING.ordinal())[3];
                        }
                        dataOut = new byte[]{b1, b2};

                        sendPacket = new DatagramPacket(dataOut, dataOut.length, ipAddress, port);
                        NetworkWriteCallBack(sendPacket);
                    } else {
                        if (this.closeAllClients) {
                            b1 = Converter.toByte(Eds3ServerMode.SERVER_SEND_SET_MODE_CMD.ordinal())[3];
                            b2 = Converter.toByte(Eds3Client.Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT.ordinal())[3];
                            dataOut = new byte[]{b1, b2};
                            wr(ip + ":" + port + ": Sending SERVER_SEND_SET_MODE_CMD msg to CLIENT_SEND_REQUEST_DROP_CLIENT");
                            sendPacket = new DatagramPacket(dataOut, dataOut.length, ipAddress, port);
                            NetworkWriteCallBack(sendPacket);
                        }
                    }
                } else {
                    b1 = Converter.toByte(Eds3ServerMode.SERVER_SEND_SET_MODE_CMD.ordinal())[3];
                    b2 = Converter.toByte(Eds3Client.Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT.ordinal())[3];
                    dataOut = new byte[]{b1, b2};
                    wr(ip + ":" + port + ": Sending SERVER_SEND_SET_MODE_CMD msg to CLIENT_SEND_REQUEST_DROP_CLIENT");
                    sendPacket = new DatagramPacket(dataOut, dataOut.length, ipAddress, port);
                    NetworkWriteCallBack(sendPacket);
                }

            } else if (dataIn[0] == Eds3ServerMode.SERVER_RECEIVE_REQUEST_OBJ_PUBLISH_L2.ordinal()) {
                wr(ip + ":" + port + ": Found SERVER_RECEIVE_REQUEST_OBJ_PUBLISH_L2 msg.");

                String version = new String(new byte[]{dataIn[1], dataIn[2], dataIn[3], dataIn[4]}, "ASCII");
                long iSentOn = Converter.toLong(new byte[]{dataIn[5], dataIn[6], dataIn[7], dataIn[8], dataIn[9], dataIn[10], dataIn[11], dataIn[12]});
                long recvOn = System.currentTimeMillis();
                MsgStats msgStats = new MsgStats(iSentOn, recvOn);
                wr(ip + ":" + port + ": XportTime: " + msgStats.diff + " ms");

                int playerId = Converter.toInt(new byte[]{dataIn[13], dataIn[14], dataIn[15], dataIn[16]});
                long iX = Converter.toLong(new byte[]{dataIn[17], dataIn[18], dataIn[19], dataIn[20], dataIn[21], dataIn[22], dataIn[23], dataIn[24]});
                long iY = Converter.toLong(new byte[]{dataIn[25], dataIn[26], dataIn[27], dataIn[28], dataIn[29], dataIn[30], dataIn[31], dataIn[32]});
                long iZ = Converter.toLong(new byte[]{dataIn[33], dataIn[34], dataIn[35], dataIn[36], dataIn[37], dataIn[38], dataIn[39], dataIn[40]});
                int objsLen = Converter.toInt(new byte[]{dataIn[41], dataIn[42], dataIn[43], dataIn[44]});
                byte[] objs = new byte[objsLen];
                
                byte[] dataOut;
                byte b1;
                byte b2;
                DatagramPacket sendPacket;
                
                try {
                    System.arraycopy(dataIn, 45, objs, 0, objs.length);
                }catch(Exception ex1) {
                    b1 = Converter.toByte(Eds3ServerMode.SERVER_SEND_SET_MODE_CMD.ordinal())[3];

                    if (this.closeAllClients) {
                        wr(ip + ":" + port + ": Sending SERVER_SEND_SET_MODE_CMD msg to CLIENT_SEND_REQUEST_DROP_CLIENT");
                        b2 = Converter.toByte(Eds3Client.Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT.ordinal())[3];
                    } else {
                        wr(ip + ":" + port + ": Sending SERVER_SEND_SET_MODE_CMD msg to WAITING");
                        b2 = Converter.toByte(Eds3Client.Eds3ClientMode.WAITING.ordinal())[3];
                    }
                    dataOut = new byte[]{b1, b2};

                    sendPacket = new DatagramPacket(dataOut, dataOut.length, ipAddress, port);
                    NetworkWriteCallBack(sendPacket);                    
                }
                    
                try {
                    r = this.players.get(ip);
                } catch (Exception ex1) {
                    this.PrintStackTrace(ex1.getStackTrace());
                    r = null;
                }

                if (r != null) {
                    p = r.players.get(ip);
                    if (p != null) {
                        p.port = port;
                        p.ipAddress = ipAddress;
                        p.objPublish = objs;

                        p.x = (float) iX;
                        p.y = (float) iY;
                        p.z = (float) iZ;

                        long start = System.currentTimeMillis();
                        p.lastCommTimeMs = iSentOn;
                        p.LogMsgStats(msgStats);
                        r.UpdateGameRoomStats();
                        r.UpdateLaggedInactivePlayers();
                        long stop = System.currentTimeMillis();
                        long diff = stop - start;
                        wr(ip + ":" + port + ": Player maintenance took: " + diff + " ms");
                    } else {
                        p = null;
                        wr(ip + ":" + port + ": Message chain dying...");
                    }
                } else {
                    p = null;
                    wr(ip + ":" + port + ": Message chain dying...");
                }

                if (r != null && p != null) {
                    if(r.inactivePlayers.size() > 0) {
                        Player iap = r.inactivePlayers.get(0);
                        GameObjectWrapperDefault iapGow = new GameObjectWrapperDefault();
                        iapGow.SetPlayerId(iap.spawnId);
                        iapGow.SetMode(GameObjectWrapperDefault.GameObjectWrapperMode.DESTROY_INACTIVE_PLAYER_CHAR.ordinal());
                        iapGow.SetId(iap.id);
                        byte[] iapB = iapGow.Serialize();
                        
                        byte[] nop = null;
                        Set<String> pks = r.players.keySet();
                        Player tiap;
                        for(String pk: pks) {
                            tiap = r.players.get(pk);                            
                            if(tiap.objPublish == null) {
                                tiap.objPublish = iapB;
                            }else {
                                nop = new byte[tiap.objPublish.length + iapB.length];
                                System.arraycopy(tiap.objPublish, 0, nop, 0, tiap.objPublish.length);
                                System.arraycopy(iapB, 0, nop, tiap.objPublish.length, iapB.length);
                                tiap.objPublish = nop;
                            }
                        }
                    }
                    
                    wr(ip + ":" + port + ": ServerVersion: " + this.version);
                    wr(ip + ":" + port + ": ClientVersion: " + version);
                    wr(ip + ":" + port + ": SpawnId/PlayerId: " + p.spawnId);
                    wr(ip + ":" + port + ": Object Id: " + p.id);
                    wr(ip + ":" + port + ": SentOn: " + iSentOn);
                    wr(ip + ":" + port + ": ReceivedOn: " + recvOn);
                    wr(ip + ":" + port + ": In SpawnId/PlayerId: " + playerId);
                    wr(ip + ":" + port + ": In X,Y,Z: " + iX + ", " + iY + ", " + iZ);
                    wr(ip + ":" + port + ": Serialized Objects Length: " + objs.length);
                    wr(ip + ":" + port + ": Serialized Object Count: " + objs.length / this.gowStatic.GetTotalLength());

                    byte[] tmsgId;
                    byte[] tmsgVer;
                    byte[] tsentOn;
                    byte[] tpId;
                    byte[] tpIp;
                    byte[] tobjs;
                    byte[] tobjsLen;
                    int totalLen;
                    byte[] lb;
                    String ttt;
                    Player tp = null;
                    dataOut = new byte[Eds3Server.DEFAULT_OUT_BUFFER];
                    int dataOutUsed = 0;
                    int dataOutStart = 0;
                    int cnt = 0;
                    Player prevTp = null;

                    for (String key : r.players.keySet()) {
                        tp = r.players.get(key);
                        if (tp != null && tp.spawnId != p.spawnId && tp.objPublish != null) {
                            prevTp = tp;
                            if (Eds3Server.OBJ_PUBLISH_L2_DIST_LIMIT_ON && Utils.SqrDist(iX, iY, iZ, tp) > Eds3Server.OBJ_PUBLISH_L2_DIST_MAX_SQR) {
                                wr(ip + ":" + port + ": Skipping msg to source player: " + tp.spawnId + " from " + p.spawnId + " too distant.");
                                continue;
                            }
                            tmsgId = new byte[]{Converter.toByte(Eds3ServerMode.SERVER_SEND_REQUESTED_OBJ_PUBLISH_L2.ordinal())[3]};
                            tmsgVer = this.version.getBytes("ASCII");
                            tsentOn = Converter.toByte(System.currentTimeMillis());
                            tpId = Converter.toByte(tp.spawnId);
                            ttt = tp.ip;
                            tpIp = ttt.getBytes("ASCII");
                            if (tpIp.length < IP_SER_LENGTH) {
                                ttt = Utils.PadRight(ttt, IP_SER_LENGTH - ttt.length());
                                tpIp = ttt.getBytes("ASCII");
                            }

                            tobjs = tp.objPublish;
                            tobjsLen = Converter.toByte(tobjs.length);

                            wr("===================tpIpLen: " + tpIp.length + " tobjsLen: " + tobjs.length + " SpawnId/PlayerId: " + tp.spawnId);
                            totalLen = tmsgId.length + tmsgVer.length + tsentOn.length + tpId.length + tpIp.length + tobjsLen.length + tobjs.length;

                            if ((dataOutUsed + totalLen) >= dataOut.length) {
                                wr(ip + ":" + port + ": Sending SERVER_SEND_REQUESTED_OBJ_PUBLISH_L2 msg for SpawnId/PlayerId: " + tp.spawnId + " FrameCount: " + cnt + " to " + p.spawnId);
                                if (this.commH != null) {
                                    dataOut = this.commH.HandleObjPublication(dataOut, dataOutUsed, this);
                                }
                                
                                if(tp != null && tp.isInactive) {
                                    wr(ip + ":" + port + ": Dropping player " + tp.spawnId + " with IP: " + tp.ip + " and ID: " + tp.id + ".");
                                    r.players.remove(tp.ip);
                                    this.players.remove(tp.ip);

                                    if(r.inactivePlayers.contains(tp)) {
                                        r.inactivePlayers.remove(tp);
                                    }

                                    if(r.lagPlayers.contains(tp)) {
                                        r.lagPlayers.remove(tp);
                                    }
                                    
                                    if(!r.freeSpawns.contains(p.spawnId)) {
                                        r.freeSpawns.add(p.spawnId);
                                    }                                    
                                }
                            
                                sendPacket = new DatagramPacket(dataOut, dataOutUsed, ipAddress, port);
                                NetworkWriteCallBack(sendPacket);

                                dataOut = new byte[Eds3Server.DEFAULT_OUT_BUFFER];
                                dataOutStart = 0;
                                dataOutUsed = 0;
                                cnt = 0;
                            }

                            dataOutStart = dataOutUsed;
                            dataOutUsed += totalLen;
                            System.arraycopy(tmsgId, 0, dataOut, dataOutStart, tmsgId.length);
                            System.arraycopy(tmsgVer, 0, dataOut, (dataOutStart + tmsgId.length), tmsgVer.length);
                            System.arraycopy(tsentOn, 0, dataOut, (dataOutStart + tmsgId.length + tmsgVer.length), tsentOn.length);
                            System.arraycopy(tpId, 0, dataOut, (dataOutStart + tmsgId.length + tmsgVer.length + tsentOn.length), tpId.length);
                            System.arraycopy(tpIp, 0, dataOut, (dataOutStart + tmsgId.length + tmsgVer.length + tsentOn.length + tpId.length), tpIp.length);
                            System.arraycopy(tobjsLen, 0, dataOut, (dataOutStart + tmsgId.length + tmsgVer.length + tsentOn.length + tpId.length + tpIp.length), tobjsLen.length);
                            System.arraycopy(tobjs, 0, dataOut, (dataOutStart + tmsgId.length + tmsgVer.length + tsentOn.length + tpId.length + tpIp.length + tobjsLen.length), tobjs.length);
                            cnt++;
                        } else {
                            if (tp == null) {
                                wr(ip + ":" + port + ": Skipping msg to player no player entry found.");
                            } else if (tp.spawnId == p.spawnId) {
                                wr(ip + ":" + port + ": Skipping msg to source player: " + tp.spawnId + ", " + p.spawnId);
                            } else if (tp.objPublish == null) {
                                wr(ip + ":" + port + ": Skipping msg to player no objects array to publish.");
                            }
                        }
                    }

                    if (dataOutUsed > 0) {
                        if (prevTp != null) {
                            wr(ip + ":" + port + ": Sending SERVER_SEND_REQUESTED_OBJ_PUBLISH_L2 msg for SpawnId/PlayerId: " + prevTp.spawnId + " to " + p.spawnId);
                        } else {
                            wr(ip + ":" + port + ": Sending SERVER_SEND_REQUESTED_OBJ_PUBLISH_L2 msg");
                        }

                        if (this.commH != null) {
                            dataOut = this.commH.HandleObjPublication(dataOut, dataOutUsed, this);
                        }

                        if(prevTp != null && prevTp.isInactive) {
                            wr(ip + ":" + port + ": Dropping player " + prevTp.spawnId + " with IP: " + prevTp.ip + " and ID: " + prevTp.id + ".");
                            r.players.remove(prevTp.ip);
                            this.players.remove(prevTp.ip);

                            if(r.inactivePlayers.contains(prevTp)) {
                                r.inactivePlayers.remove(prevTp);
                            }

                            if(r.lagPlayers.contains(prevTp)) {
                                r.lagPlayers.remove(prevTp);
                            }
                            
                            if (!r.freeSpawns.contains(p.spawnId))
                            {
                                r.freeSpawns.add(p.spawnId);
                            }                            
                        }

                        sendPacket = new DatagramPacket(dataOut, dataOutUsed, ipAddress, port);
                        NetworkWriteCallBack(sendPacket);

                        dataOut = new byte[Eds3Server.DEFAULT_OUT_BUFFER];
                        dataOutStart = 0;
                        dataOutUsed = 0;
                    }

                    if (this.commH != null) {
                        lb = this.commH.HandleProcessServerGame(System.currentTimeMillis(), this);
                        if (lb != null && lb.length > 0) {
                            wr(ip + ":" + port + ": Sending HandleProcessServerGame msg");
                            sendPacket = new DatagramPacket(lb, lb.length, ipAddress, port);
                            NetworkWriteCallBack(sendPacket);
                        }
                    }

                    if (SERVER_DRIVEN_MODE_ON) {
                        b1 = Converter.toByte(Eds3ServerMode.SERVER_SEND_SET_MODE_CMD.ordinal())[3];

                        if (this.closeAllClients) {
                            wr(ip + ":" + port + ": Sending SERVER_SEND_SET_MODE_CMD msg to CLIENT_SEND_REQUEST_DROP_CLIENT");
                            b2 = Converter.toByte(Eds3Client.Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT.ordinal())[3];
                        } else {
                            wr(ip + ":" + port + ": Sending SERVER_SEND_SET_MODE_CMD msg to WAITING");
                            b2 = Converter.toByte(Eds3Client.Eds3ClientMode.WAITING.ordinal())[3];
                        }
                        dataOut = new byte[]{b1, b2};

                        sendPacket = new DatagramPacket(dataOut, dataOut.length, ipAddress, port);
                        NetworkWriteCallBack(sendPacket);
                    } else {
                        if (this.closeAllClients) {
                            b1 = Converter.toByte(Eds3ServerMode.SERVER_SEND_SET_MODE_CMD.ordinal())[3];
                            b2 = Converter.toByte(Eds3Client.Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT.ordinal())[3];
                            dataOut = new byte[]{b1, b2};
                            wr(ip + ":" + port + ": Sending SERVER_SEND_SET_MODE_CMD msg to CLIENT_SEND_REQUEST_DROP_CLIENT");
                            sendPacket = new DatagramPacket(dataOut, dataOut.length, ipAddress, port);
                            NetworkWriteCallBack(sendPacket);
                        }
                    }
                } else {
                    b1 = Converter.toByte(Eds3ServerMode.SERVER_SEND_SET_MODE_CMD.ordinal())[3];
                    b2 = Converter.toByte(Eds3Client.Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT.ordinal())[3];
                    dataOut = new byte[]{b1, b2};
                    wr(ip + ":" + port + ": Sending SERVER_SEND_SET_MODE_CMD msg to CLIENT_SEND_REQUEST_DROP_CLIENT");
                    sendPacket = new DatagramPacket(dataOut, dataOut.length, ipAddress, port);
                    NetworkWriteCallBack(sendPacket);
                }

            } else {
                if (this.commH != null) {
                    this.commH.HandleProcessModeUnknown(dataIn, this);
                }
            }

        } catch (Exception e) {
            if (!this.shuttingDown) {
                wr(e.getMessage());
                PrintStackTrace(e.getStackTrace());
            }
        }
    }

    public synchronized void NetworkWriteCallBack(DatagramPacket sendPacket) {
        //Retrieve the state object and the handler socket.
        wr("");
        wr("");
        wr("NetworkWriteCallBack");

        try {
            int tlen = sendPacket.getLength();
            wr("DataOutLen: " + tlen + " TLen: " + tlen);

            if (this.shuttingDown || this.serverSocket.isClosed()) {
                return;
            }

            this.totalDataOut += tlen;
            this.totalMsgsOut++;
            serverSocket.send(sendPacket);
        } catch (Exception e) {
            wr(e.getMessage());
            PrintStackTrace(e.getStackTrace());
        }
    }
        
    public void SetDebugOn(boolean b) {
        this.debugOn = b;
    }
    //</editor-fold>
}
