package com.middlemind.Eds3;

import java.util.ArrayList;
import java.util.Hashtable;

/**
 *
 * @author Victor G. Brusca, Middlemind Games
 */
public class GameRoom {

    public enum GameRoomMode {
        NO_STATE,
        IN_LOBBY,
        IN_GAME,
        STARTING_GAME,
        STOPPING_GAME,
        SPECIAL_1,
        SPECIAL_2,
        SPECIAL_3,
        SPECIAL_4,
        SPECIAL_5,
        SPECIAL_6,
        SPECIAL_7,
        SPECIAL_8,
        SPECIAL_9,
        SPECIAL_10,
        SERVER_SHUTTING_DOWN
    }

    public static double PLAYER_THRESHOLD_LAG_PRCT = 0.25;
    public static long PLAYER_THRESHOLD_INACTIVE_COMM_MS = 10000;

    private int spawnId = 0;
    public int id = -1;
    public Hashtable<String, Player> players = null;
    public GameRoomMode mode;
    public int dunGenSeed = 0;
    public String serverVersion = null;
    public ArrayList<Player> lagPlayers = null;
    public ArrayList<Player> inactivePlayers = null;
    public double msgAvgTimeMs = 0.0;
    public int msgTotalCnt = 0;
    public long msgTotalXportTimeMs = 0;
    public ArrayList<Integer> freeSpawns = null;
    public Eds3Server eds3server = null;

    @SuppressWarnings("Convert2Diamond")
    public GameRoom() {
        this.players = new Hashtable<String, Player>();
        this.lagPlayers = new ArrayList<Player>();
        this.inactivePlayers = new ArrayList<Player>();
        this.mode = GameRoomMode.IN_LOBBY;
        this.freeSpawns = new ArrayList<Integer>();
    }

    @SuppressWarnings("UnusedAssignment")
    public int GetSpawnId() {
        int ret = 0;
        if (this.freeSpawns.size() > 0) {
            ret = this.freeSpawns.remove(0);
        } else {
            this.spawnId++;
            ret = this.spawnId;
        }
        return ret;
    }

    @SuppressWarnings("UnusedAssignment")
    public void UpdateGameRoomStats() {
        int cnt = 0;
        double avg = 0.0;
        long totalDiff = 0;
        Player p = null;
        for (String key : this.players.keySet()) {
            p = this.players.get(key);
            cnt += p.msgTotalCnt;
            totalDiff += p.msgTotalXportTimeMs;
        }

        if (cnt > 0) {
            avg = (double) totalDiff / (double) cnt;
        } else {
            avg = 0.0;
        }

        this.msgAvgTimeMs = avg;
        this.msgTotalCnt = cnt;
        this.msgTotalXportTimeMs = totalDiff;
    }

    @SuppressWarnings("UnusedAssignment")
    public void UpdateLaggedInactivePlayers() {
        Player p = null;
        long now = System.currentTimeMillis();

        for (String key : this.players.keySet()) {
            p = this.players.get(key);
            double prctDiff = (p.msgAvgTimeMs - this.msgAvgTimeMs) / this.msgAvgTimeMs;

            Utils.wr("Inactive Player Time Calc: " + (now - p.lastCommTimeMs));
            if (now - p.lastCommTimeMs > PLAYER_THRESHOLD_INACTIVE_COMM_MS) {
                p.isInactive = true;
                p.isLagging = false;
                
                if(this.inactivePlayers.contains(p) != true) {
                    this.inactivePlayers.add(p);    
                }else if(this.inactivePlayers.contains(p) && (now - p.lastCommTimeMs) > PLAYER_THRESHOLD_INACTIVE_COMM_MS * 2) {
                    Utils.wr("Drop from game...");
                    if(this.eds3server.GetPlayers().containsKey(p.ip)) {
                        this.eds3server.GetPlayers().remove(p.ip);
                    }
                    this.inactivePlayers.remove(p);
                    this.players.remove(p.ip);
                    if(!freeSpawns.contains(p.spawnId)) {
                        freeSpawns.add(p.spawnId);
                    }                    
                }
                
                if(this.lagPlayers.contains(p)) {
                    this.lagPlayers.remove(p);
                }
            } else if (prctDiff > 0 && prctDiff > PLAYER_THRESHOLD_LAG_PRCT) {
                p.isLagging = true;
                p.isInactive = false;
                
                if(this.lagPlayers.contains(p) != true) {
                    this.lagPlayers.add(p);
                }
                
                if(this.inactivePlayers.contains(p)) {
                    this.inactivePlayers.remove(p);
                }
            } else {
                p.isInactive = false;
                p.isLagging = false;
            }
        }
    }

    @Override
    public String toString() {
        String ret = "";
        ret += "CurrentSpawnId: " + this.spawnId + "\n";
        ret += "Id: " + this.id + "\n";
        ret += "PlayerCount: " + this.players.size() + "\n";
        ret += "CurrentMode: " + this.mode + "\n";
        ret += "DunGenSeed: " + this.dunGenSeed + "\n";
        ret += "ServerVersion: " + this.serverVersion + "\n";
        ret += "LagPlayerCount: " + this.lagPlayers.size() + "\n";
        ret += "MsgAvgTimeMs: " + this.msgAvgTimeMs + "\n";
        ret += "MsgTotalCnt: " + this.msgTotalCnt + "\n";
        ret += "MsgTotalXportTimeMs: " + this.msgTotalXportTimeMs + "\n";
        return ret;
    }
}
