package com.middlemind.Eds3;

import java.net.InetAddress;
import java.util.ArrayList;

/**
 *
 * @author Victor G. Brusca, Middlemind Games
 */
public class Player implements Comparable<Player> {

    public static final int MAX_MSG_STATS = 100;
    public int id = -1;
    public int roomId = -1;
    public String ip = null;
    public int spawnId = -1;
    public int port = -1;
    public InetAddress ipAddress = null;
    public byte[] objPublish = null;
    public String clientVersion = null;
    public ArrayList<MsgStats> msgStats = null;
    public double msgAvgTimeMs = 0;
    public int msgTotalCnt = 0;
    public long msgTotalXportTimeMs = 0;
    public long lastCommTimeMs = 0;
    public boolean isInactive = false;
    public boolean isLagging = false;
    public float x;
    public float y;
    public float z;

    @SuppressWarnings("Convert2Diamond")
    public Player(String Ip) {
        this.x = 0.0f;
        this.y = 0.0f;
        this.z = 0.0f;
        this.msgStats = new ArrayList<MsgStats>();
        this.ip = Ip;
    }

    @SuppressWarnings("Convert2Diamond")
    public Player(int Id, int RoomId, String Ip) {
        this.x = 0.0f;
        this.y = 0.0f;
        this.z = 0.0f;
        this.msgStats = new ArrayList<MsgStats>();
        this.id = Id;
        this.roomId = RoomId;
        this.ip = Ip;
    }

    @SuppressWarnings("Convert2Diamond")
    public Player(int Id, int RoomId, String Ip, int Port) {
        this.x = 0.0f;
        this.y = 0.0f;
        this.z = 0.0f;
        this.msgStats = new ArrayList<MsgStats>();
        this.id = Id;
        this.roomId = RoomId;
        this.ip = Ip;
        this.port = Port;
    }

    public void LogMsgStats(MsgStats stats) {
        if (this.msgStats.size() > MAX_MSG_STATS) {
            this.msgStats.remove(0);
        }
        this.msgStats.add(stats);
        UpdateMsgStats();
    }

    @SuppressWarnings("UnusedAssignment")
    private void UpdateMsgStats() {
        int len = this.msgStats.size();
        int cnt = 0;
        double avg = 0.0;
        long totalDiff = 0;
        for (int i = 0; i < len; i++) {
            cnt++;
            totalDiff += this.msgStats.get(i).diff;
        }

        if (cnt > 0) {
            avg = (double) totalDiff / (double) cnt;
        } else {
            avg = 0.0;
        }

        this.msgAvgTimeMs = avg;
        this.msgTotalCnt = cnt;
        this.msgTotalXportTimeMs = totalDiff;
    }

    @Override
    public String toString() {
        String ret = "";
        ret += "Id: " + this.id + "\n";
        ret += "RoomId: " + this.roomId + "\n";
        ret += "IP: " + this.ip + "\n";
        ret += "SpawnId/PlayerId: " + this.spawnId + "\n";
        ret += "Port: " + this.port + "\n";
        ret += "ClientVersion: " + this.clientVersion + "\n";
        ret += "MsgStatsCount: " + this.msgStats.size() + "\n";
        ret += "MsgAvgTimeMs: " + this.msgAvgTimeMs + "\n";
        ret += "MsgTotalCnt: " + this.msgTotalCnt + "\n";
        ret += "MsgTotalXportTimeMs: " + this.msgTotalXportTimeMs + "\n";
        return ret;
    }

    @Override
    public int compareTo(Player p) {
        return Double.compare(this.msgAvgTimeMs, p.msgAvgTimeMs);
    }
}
