package com.middlemind.Eds3.mock;

import com.middlemind.Eds3.Utils;
import com.middlemind.Eds3.unity.GameObjectManager;
import com.middlemind.Eds3.unity.GameObjectManagerDefault;
import com.middlemind.Eds3.unity.GameObjectWrapperDefault;
import com.middlemind.Eds3.unity.GameObjectWrapper;

/**
 *
 * @author Victor G. Brusca, Middlemind Games
 */
public class MockGenDungeon {

    private GameObjectManager man;
    private int dunGenSeed;
    private int playerId;              //also the spawnId
    public static boolean SLEEP = false;
    public static boolean RANDOM = false;
    public static boolean MOCK_PLAYER_CHAR = true;
    
    public MockGenDungeon(int seed) {
        this.dunGenSeed = seed;
        this.playerId = 0;
        this.man = new GameObjectManagerDefault();
    }    
    
    public MockGenDungeon(int seed, int spawnId) {
        this.dunGenSeed = seed;
        this.playerId = spawnId;
        this.man = new GameObjectManagerDefault();
    }    
    
    public void GenerateObjects() {
        int min = 30;
        int max = 60;
        int objs = RANDOM ? (int)(Math.random() * (max - min) + min) : min;
        int nObjs = RANDOM ? (int)(Math.random() * (max - min) + min) : min;
        int i = 0;
        
        Utils.wr("Generating " + objs + " game objects.");
        for(i = 0; i < objs; i++) {
            this.man.AddGameObject(new GameObjectWrapperDefault(false, playerId), false);
        }
        
        Utils.wr("Generating " + nObjs + " network game objects.");
        for(i = 0; i < nObjs; i++) {
            this.man.AddGameObject(new GameObjectWrapperDefault(true, playerId), true);
        }
        
        if (MOCK_PLAYER_CHAR) {
            this.man.SetPlayerChar(new GameObjectWrapperDefault(true, playerId));
        }
        
        if(SLEEP) {
            try {
                Thread.sleep(5000);

            }catch(Exception e) {
                e.printStackTrace();            
            }
        }
    }
    
    public void GenerateSpawnObjects() {
        int min = 30;
        int max = 60;
        int nObjs = RANDOM ? (int)(Math.random() * (max - min) + min) : min;
        int i = 0;
                
        Utils.wr("Generating spawn " + nObjs + " spawn objects.");
        for(i = 0; i < nObjs; i++) {
            this.man.AddGameObject(new GameObjectWrapperDefault(true), true);
        }
        
        if(SLEEP) {        
            try {
                Thread.sleep(5000);

            }catch(Exception e) {
                e.printStackTrace();            
            }
        }
    }
    
    public void RandomlyAdjustObjects() {
        int min = 10;
        int max = 20;
        int objs = RANDOM ? (int)(Math.random() * (max - min) + min) : min;
        int nObjs = RANDOM ? (int)(Math.random() * (max - min) + min) : min;
        int i = 0;
        
        Utils.wr("Adjusting " + objs + " game objects.");
        for(i = 0; i < objs; i++) {
            this.man.GetGameObjects().remove(Utils.GetRandomKey(this.man.GetGameObjects()));
            this.man.AddGameObject(new GameObjectWrapperDefault(false), false);
        }
        
        Utils.wr("Adjusting " + nObjs + " network game objects.");
        for(i = 0; i < nObjs; i++) {
            this.man.GetNetworkGameObjects().remove(Utils.GetRandomKey(this.man.GetNetworkGameObjects()));
            this.man.AddGameObject(new GameObjectWrapperDefault(true), true);
        }
             
        Utils.wr("Found " + this.man.GetGameObjects().size() + " gameObjects.");
        Utils.wr("Found " + this.man.GetNetworkGameObjects().size() + " networkGameObjects.");
    }
    
    public byte[] SerializeNetworkGameObjects() {
        return this.man.SerializeNetworkGameObjects();
    }
    
    public void DeserializeNetworkGameObjects(byte[] bIn) {
        this.man.DeserializeNetworkGameObjects(bIn);
    }
    
    public byte[] SerializeGameObjects() {
        return this.man.SerializeGameObjects();
    }
    
    public void DeserializeGameObjects(byte[] bIn) {
        this.man.DeserializeGameObjects(bIn);
    }

    public boolean NetworkGameObjectsEquals(MockGenDungeon mgd) {
        if(mgd != null) {
            if(mgd.dunGenSeed == this.dunGenSeed && this.man.NetworkGameObjectsEquals(mgd.man)) {
                return true;
            }else {
                return false;
            }
        }else {
            return false;
        }
    }
    
    public boolean GameObjectsEquals(MockGenDungeon mgd) {
        if(mgd != null) {
            if(mgd.dunGenSeed == this.dunGenSeed && this.man.GameObjectsEquals(mgd.man)) {
                return true;
            }else {
                return false;
            }
        }else {
            return false;
        }
    }

    public int GetGameObjectsCount() {
        return this.man.GetGameObjects().size();
    }
    
    public int GetNetworkGameObjectsCount() {
        return this.man.GetNetworkGameObjects().size();
    }

    public GameObjectManager GetGameObjectManager() {
        return this.man;
    }
    
    public GameObjectWrapper GetPlayerChar() {
        return this.man.GetPlayerChar();
    }
}
