package com.middlemind.Eds3.unity;

import com.middlemind.Eds3.HandleGameObjectWrapperSerialization;
import com.middlemind.Eds3.Utils;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Set;

/**
 *
 * @author Victor G. Brusca, Middlemind Games
 */
public class GameObjectManagerDefault implements GameObjectManager {

    private Hashtable<Integer, GameObjectWrapper> gameObjects;
    private Hashtable<Integer, GameObjectWrapper> networkGameObjects;
    private Hashtable<Integer, GameObjectWrapper> inNetworkGameObjects;    
    private HandleGameObjectWrapperSerialization serH = null;
    private GameObjectWrapper playerChar = null;
    private Hashtable<String, Integer> nameToGameObjectId = null;

    public GameObjectManagerDefault() {
        this.playerChar = new GameObjectWrapperDefault();
        this.gameObjects = new Hashtable();
        this.networkGameObjects = new Hashtable();
        this.nameToGameObjectId = new Hashtable();
        this.inNetworkGameObjects = new Hashtable();
    }

    @Override
    public Hashtable<Integer, GameObjectWrapper> GetGameObjects() {
        return this.gameObjects;
    }

    @Override
    public Hashtable<Integer, GameObjectWrapper> GetNetworkGameObjects() {
        return this.networkGameObjects;
    }
    
    @Override
    public Hashtable<Integer, GameObjectWrapper> GetInNetworkGameObjects() {
        return this.inNetworkGameObjects;
    }
    
    @Override
    public Hashtable<String, Integer> GetNameToGameObjectId() {
        return this.nameToGameObjectId;
    }
        
    @Override
    public GameObjectWrapper GetPlayerChar() {
        return this.playerChar;
    }

    @Override
    public void SetPlayerChar(GameObjectWrapper pc) {
        this.playerChar = pc;
    }

    @Override
    public HandleGameObjectWrapperSerialization GetSerializationHandler() {
        return this.serH;
    }

    @Override
    public void SetSerializationHandler(HandleGameObjectWrapperSerialization SerH) {
        this.serH = SerH;
    }

    @Override
    public void RegisterGameObjectName(String name, int id) {
        if (this.nameToGameObjectId.containsKey(name)) {
            this.nameToGameObjectId.remove(name);
        }
        this.nameToGameObjectId.put(name, id);
    }

    @Override
    public void AddGameObject(GameObjectWrapper obj, boolean isNetwork) {
        if (isNetwork) {
            if (this.networkGameObjects.containsKey(obj.GetId())) {
                this.networkGameObjects.remove(obj.GetId());
            }

            this.networkGameObjects.put(obj.GetId(), obj);
        } else {
            if (this.gameObjects.containsKey(obj.GetId())) {
                this.gameObjects.remove(obj.GetId());
            }

            this.gameObjects.put(obj.GetId(), obj);
        }
    }

    @Override
    public void AddInNetworkGameObject(GameObjectWrapper obj) {
        if (this.inNetworkGameObjects.containsKey(obj.GetId())) {
            this.inNetworkGameObjects.remove(obj.GetId());
        }

        this.inNetworkGameObjects.put(obj.GetId(), obj);
    }
    
    @Override
    public byte[] SerializeNetworkGameObjects() {
        return this.SerializeDirect(this.networkGameObjects);
    }

    @Override
    public byte[] SerializeGameObjects() {
        return this.SerializeDirect(this.gameObjects);
    }

    public byte[] Serialize(Hashtable<Integer, GameObjectWrapper> bIn) {
        ArrayList<GameObjectWrapper> bOut = new ArrayList();
        bOut.addAll(bIn.values());
        return this.Serialize(bOut);
    }

    public byte[] Serialize(ArrayList<GameObjectWrapper> bIn2) {
        ArrayList<GameObjectWrapper> bIn = new ArrayList<GameObjectWrapper>(bIn2);
        int len = bIn.size();

        boolean hasPlayerChar = false;
        if (this.playerChar != null) {
            hasPlayerChar = true;
            len += 1;
        }

        int off = 0;
        int[] offs = new int[len + 1];
        int[] lengths = new int[len];

        ArrayList<byte[]> pRet = new ArrayList();
        int totalLen = 0;
        byte[] b;
        byte[] ret;
        GameObjectWrapper gow;

        for (int i = 0; i < len; i++) {
            if (hasPlayerChar && i == 0) {
                gow = this.playerChar;

            } else {
                gow = bIn.get(i);

            }

            if (this.serH != null) {
                gow = this.serH.BeforeSerializeGow(gow);
            }

            b = gow.Serialize();

            if (this.serH != null) {
                b = this.serH.AfterSerializeGow(b, gow);
            }

            pRet.add(b);

            if (this.serH != null) {
                this.serH.SerializeIteration(gow, b, pRet, off);
            }

            offs[i] = off;
            off += b.length;
            lengths[i] = b.length;
            totalLen += b.length;
        }
        offs[offs.length - 1] = totalLen;

        ret = new byte[totalLen];
        for (int i = 0; i < len; i++) {
            System.arraycopy(pRet.get(i), 0, ret, offs[i], lengths[i]);
        }

        return ret;
    }

    public byte[] SerializeDirect(Hashtable<Integer, GameObjectWrapper> bIn2) {
        Hashtable<Integer, GameObjectWrapper> bIn = new Hashtable<Integer, GameObjectWrapper>(bIn2);
        int len = bIn.size();

        boolean hasPlayerChar = false;
        if (this.playerChar != null) {
            hasPlayerChar = true;
            len += 1;
        }

        int off = 0;
        int totalLen = len * GameObjectWrapperDefault.TOTAL_LENGTH;
        byte[] b;
        byte[] ret = new byte[totalLen];
        GameObjectWrapper gow2;

        if (hasPlayerChar) {
            gow2 = this.playerChar;

            if (this.serH != null) {
                gow2 = this.serH.BeforeSerializeGow(gow2);
            }

            b = gow2.Serialize();

            if (this.serH != null) {
                b = this.serH.AfterSerializeGow(b, gow2);
            }

            System.arraycopy(b, 0, ret, off, GameObjectWrapperDefault.TOTAL_LENGTH);

            if (this.serH != null) {
                this.serH.SerializeIteration(gow2, b, ret, off);
            }

            off += b.length;
        }

        for (GameObjectWrapper gow : bIn.values()) {
            gow2 = gow;

            if (this.serH != null) {
                gow2 = this.serH.BeforeSerializeGow(gow2);
            }

            b = gow2.Serialize();

            if (this.serH != null) {
                b = this.serH.AfterSerializeGow(b, gow2);
            }

            System.arraycopy(b, 0, ret, off, GameObjectWrapperDefault.TOTAL_LENGTH);

            if (this.serH != null) {
                this.serH.SerializeIteration(gow2, b, ret, off);
            }

            off += b.length;
        }
        return ret;
    }

    public ArrayList<GameObjectWrapper> Deserialize2ArrayList(byte[] bIn) {
        int len = bIn.length;
        ArrayList<GameObjectWrapper> ret = new ArrayList();
        GameObjectWrapper gow = null;
        int msgLen = GameObjectWrapperDefault.TOTAL_LENGTH;
        byte[] d = null;
        int pos = 0;

        if (this.playerChar != null) {
            d = new byte[msgLen];
            System.arraycopy(bIn, pos, d, 0, msgLen);
            gow = new GameObjectWrapperDefault();

            if (this.serH != null) {
                d = this.serH.BeforeDeserializeGow(d, gow);
            }

            gow.Deserialize(d);

            if (this.serH != null) {
                gow = this.serH.AfterDeserializeGow(gow);
            }

            this.playerChar = gow;

            if (this.serH != null) {
                this.serH.DeserializeIteration(gow, d, ret, pos);
            }

            pos += msgLen;
            len -= msgLen;
        }

        while (len >= msgLen) {
            d = new byte[msgLen];
            System.arraycopy(bIn, pos, d, 0, msgLen);
            gow = new GameObjectWrapperDefault();

            if (this.serH != null) {
                d = this.serH.BeforeDeserializeGow(d, gow);
            }

            gow.Deserialize(d);

            if (this.serH != null) {
                gow = this.serH.AfterDeserializeGow(gow);
            }

            ret.add(gow);

            if (this.serH != null) {
                this.serH.DeserializeIteration(gow, d, ret, pos);
            }

            pos += msgLen;
            len -= msgLen;
        }

        Utils.wr("Found " + ret.size() + " game objects.");
        return ret;
    }

    public Hashtable<Integer, GameObjectWrapper> Deserialize2Hashtable(byte[] bIn) {
        ArrayList<GameObjectWrapper> tmp = this.Deserialize2ArrayList(bIn);
        GameObjectWrapper gow = null;
        int len = tmp.size();
        Hashtable<Integer, GameObjectWrapper> ret = new Hashtable();

        for (int i = 0; i < len; i++) {
            gow = tmp.get(i);
            ret.put(gow.GetId(), gow);
        }

        Utils.wr("Found " + ret.size() + " game objects.");
        return ret;
    }

    public Hashtable<Integer, GameObjectWrapper> Deserialize2HashtableDirect(byte[] bIn) {
        int len = bIn.length;
        Hashtable<Integer, GameObjectWrapper> ret = new Hashtable();
        GameObjectWrapper gow = null;
        int msgLen = GameObjectWrapperDefault.TOTAL_LENGTH;
        byte[] d = null;
        int pos = 0;

        if (this.playerChar != null) {
            d = new byte[msgLen];
            System.arraycopy(bIn, pos, d, 0, msgLen);
            gow = new GameObjectWrapperDefault();

            if (this.serH != null) {
                d = this.serH.BeforeDeserializeGow(d, gow);
            }

            gow.Deserialize(d);

            if (this.serH != null) {
                gow = this.serH.AfterDeserializeGow(gow);
            }

            this.playerChar = gow;

            if (this.serH != null) {
                this.serH.DeserializeIteration(gow, d, ret, pos);
            }

            pos += msgLen;
            len -= msgLen;
        }

        while (len >= msgLen) {
            d = new byte[msgLen];
            System.arraycopy(bIn, pos, d, 0, msgLen);
            gow = new GameObjectWrapperDefault();

            if (this.serH != null) {
                d = this.serH.BeforeDeserializeGow(d, gow);
            }

            gow.Deserialize(d);

            if (this.serH != null) {
                gow = this.serH.AfterDeserializeGow(gow);
            }

            ret.put(gow.GetId(), gow);

            if (this.serH != null) {
                this.serH.DeserializeIteration(gow, d, ret, pos);
            }

            pos += msgLen;
            len -= msgLen;
        }

        Utils.wr("Found " + ret.size() + " game objects.");
        return ret;
    }

    @Override
    public void DeserializeNetworkGameObjects(byte[] bIn) {
        this.networkGameObjects = this.Deserialize2HashtableDirect(bIn);
    }

    @Override
    public void DeserializeGameObjects(byte[] bIn) {
        this.gameObjects = this.Deserialize2HashtableDirect(bIn);
    }

    @Override
    public boolean NetworkGameObjectsEquals(GameObjectManager gom) {
        Set<Integer> keys = gom.GetNetworkGameObjects().keySet();
        GameObjectWrapper gow1 = null;
        GameObjectWrapper gow2 = null;

        for (Integer key : keys) {
            gow1 = gom.GetNetworkGameObjects().get(key);
            gow2 = this.networkGameObjects.get(key);

            if (gow1.Equals(gow2) == false) {
                return false;
            }
        }

        return true;
    }

    @Override
    public boolean GameObjectsEquals(GameObjectManager gom) {
        Set<Integer> keys = gom.GetGameObjects().keySet();
        GameObjectWrapper gow1 = null;
        GameObjectWrapper gow2 = null;

        for (Integer key : keys) {
            gow1 = gom.GetGameObjects().get(key);
            gow2 = this.gameObjects.get(key);

            if (gow1.Equals(gow2) == false) {
                return false;
            }
        }

        return true;
    }
}
